#!/usr/bin/python
#transport_email.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2013
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#


import os
import sys
import smtplib
import re
import time
import tempfile
from StringIO import StringIO


##import email
import email.Encoders
import email.MIMEMultipart
import email.MIMEBase


try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in transport_email.py')


import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)
from twisted.mail.smtp import sendmail
from twisted.mail.smtp import SMTPSenderFactory
from twisted.mail.smtp import ESMTPSenderFactory
from twisted.mail.pop3 import AdvancedPOP3Client as POP3Client


from twisted.internet.defer import Deferred, succeed
from twisted.internet.protocol import ClientFactory
from twisted.internet.ssl import ClientContextFactory
from OpenSSL.SSL import SSLv3_METHOD


import misc
import dhnio
import settings
import transport_control
import tmpfile

#info: [email, server, port, username, password, ssl, need_auth]

##misc.DebugLevel = 14

log_state = False

#maximum length of any single line in the message
MAX_LINE_LENGTH = 1024

ready_receiving = True

#-------------------------------------------------------------------------------

PollingObject = None
class PollingControl:
    timemark = time.time()
    #LAST_ACTIVITY_PERIOD = 15.0*60.0   #last activity period. then increase polling time

    current_polling_interval = 30.0  # start off checking every minute
    SHORT_POLLING = 10.0   # if we had activity recently, check every 15 seconds
    MEDIUM_POLLING = 30.0  # after 1 minute of inactivity check every minute until 15 minutes
    LONG_POLLING = 90.0   # after 15 minutes of inactivity check every 15 minutes

    def init(self):
        self.clear()
        self.update()

    def clear(self):
        self.timemark = time.time()

    def deltatime(self):
        return time.time() - self.timemark

# we don't want to update the email timeout based on activity
#    def update_timeout(self):
#        if self.deltatime() < self.LAST_ACTIVITY_PERIOD:
#            misc.setEmailSendTimeout(settings.SendTimeOutEmail())
#        else:
#            misc.setEmailSendTimeout(settings.SendTimeOutEmail() * 10)
##        transport_control.log('email', 'use % timeout' % (str(misc.getEmailSendTimeout())))

#
#    def update_polltime(self):
#        if self.deltatime() < self.LAST_ACTIVITY_PERIOD:
#            misc.setEmailPollingTime(settings.EmailPollingTime())
#        else:
#            misc.setEmailPollingTime(settings.EmailPollingTime() * 10)
##        transport_control.log('email', 'use %s polling time' % (str(misc.getEmailPollingTime())))

    def update_polltime(self):
        if self.deltatime() > self.LONG_POLLING:
            self.current_polling_interval = self.LONG_POLLING
        elif self.deltatime() > self.MEDIUM_POLLING:
            self.current_polling_interval = self.MEDIUM_POLLING
        else:
            self.current_polling_interval = self.SHORT_POLLING

    def update(self):
        dhnio.Dprint(14, 'transport_email:PoolingControl.update: deltatime='+str(self.deltatime()))
        self.update_polltime()
        #self.update_timeout()

    def receive(self):
##        self.counter += 1.0
##        if self.counter > self.MAX_COUNTER:
        self.update()
        self.clear()

    def send(self):
        self.update()
        self.clear()


def check_info(info):
    if len(info) < 6:
        return False
    if info[0].strip() == '' or info[1].strip() == '':
        return False
    try:
        port_ = int(info[2])
    except:
        return False
    return True


def poll():
    global PollingObject
    if PollingObject is None:
        PollingObject = PollingControl()
    return PollingObject

def pollingInterval():
    return poll().current_polling_interval

#-------------------------------------------------------------------------------

##############################################################
## Sending ###################################################
##############################################################

def send_smtplib(host, login, password, ssl_flag, addr, to, subject, filename):
    dhnio.Dprint(14, 'transport_email.send_smtplib to: ' + str(to))

    fin = file(filename, 'rb')
    bin_data = fin.read()
    fin.close()



    msg = email.MIMEMultipart.MIMEMultipart()
    msg["Subject"] = subject
    msg["From"]    = addr
    msg["To"]      = to
    part = email.MIMEBase.MIMEBase('application', "octet-stream")
    part.set_payload( bin_data+'end' )
    email.Encoders.encode_base64(part)
    part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(filename))
    msg.attach(part)

    try:
        server = smtplib.SMTP(host)
    except:
##        transport_control.msg('e', 'could not conect to smtp server')
##        transport_control.connectionStatusCallback('email', to, None, 'could not conect to smtp server')
        transport_control.sendStatusReport(to, filename,
        'failed', 'email',
        None, 'could not conect to smtp server')
        return

    if dhnio.DebugLevel > 14:
        server.set_debuglevel(1)
    dhnio.Dprint(14, 'transport_email.send_smtplib: SSL FLAG:'+str(ssl))
    if ssl_flag:
        try:
            dhnio.Dprint(14, 'transport_email.send_smtplib ehlo')
            server.ehlo()
            dhnio.Dprint(14, 'transport_email.send_smtplib starttls')
            server.starttls()
            dhnio.Dprint(14, 'transport_email.send_smtplib ehlo')
            server.ehlo()
        except:
            dhnio.Dprint(14, 'transport_email.send_smtplib: No support ssl')
##            transport_control.connectionStatusCallback('email', to, None, 'server did not support ssl' )
            transport_control.sendStatusReport(to, filename,
            'failed', 'email',
            None, 'server did not support ssl')
##            transport_control.msg('e', 'server did not support ssl')

    try:
        dhnio.Dprint(14, 'transport_email.send_smtplib login')
        server.login(login, password)
    except:
        dhnio.Dprint(14, 'transport_email.send_smtplib: Not need to login')

    try:
        dhnio.Dprint(14, 'transport_email.send_smtplib: sendmail')
        server.sendmail(addr, [to], msg.as_string())
        dhnio.Dprint(14, 'transport_email.send_smtplib: close')
        server.close()
        dhnio.Dprint(14, 'transport_email.send_smtplib: done')
    except:
##        transport_control.connectionStatusCallback('email', to, None, 'unknown error during sending message')
        transport_control.sendStatusReport(to, filename,
        'failed', 'email',
        None, 'unknown error during sending message')
##        transport_control.msg('e', 'can not send message')

def sendTwistedMail(host, addr, to, subject, filename, port=25):
    dhnio.Dprint(14, 'transport_email.sendTwistedMail to: ' + str(to))

    fin = file(filename, 'rb')
    bin_data = fin.read()
    fin.close()

    msg = email.MIMEMultipart.MIMEMultipart()
    msg["Subject"] = subject
    msg["From"]    = addr
    msg["To"]      = to

    part = email.MIMEBase.MIMEBase('application', "octet-stream")
    part.set_payload( bin_data+'end' )
    email.Encoders.encode_base64(part)
    part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(filename))
    msg.attach(part)

    try:
        port_ = int(port)
    except:
        port_ = 25

    return sendmail(host, addr, [to], msg.as_string(), port=port_)

class MySenderFactory(ESMTPSenderFactory):
    pass
##    def buildProtocol(self, addr):
##        p = ESMTPSenderFactory.buildProtocol(self, addr)
##        dhnio.Dprint(6, 'transport_email.MyESMTPSenderFactory.buildProtocol obj=%s tls=%s' % (str(p), str(p.tlsMode)))
##        return p


def sendTwistedMailAuth(host, port, login, password, ssl_flag, need_login, addr, to, subject, filename):
    dhnio.Dprint(14, 'transport_email.sendTwistedMailAuth to:%s ssl:%s auth:%s ' % (str(to),str(ssl_flag), str(need_login)))

    fin = file(filename, 'rb')
    bin_data = fin.read()
    fin.close()

    msg = email.MIMEMultipart.MIMEMultipart()
    msg["Subject"] = subject
    msg["From"]    = addr
    msg["To"]      = to

    part = email.MIMEBase.MIMEBase('application', "octet-stream")
    part.set_payload( bin_data+'end' )
    email.Encoders.encode_base64(part)
    part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(filename))
    msg.attach(part)


    try:
        port_ = int(port)
    except:
        port_ = 25
    messageText = msg.as_string()
    cf = ClientContextFactory()
    cf.method = SSLv3_METHOD
    result = Deferred()
    factory = MySenderFactory(
        login,
        password,
        addr,
        to,
        StringIO(messageText),
        result,
        contextFactory=cf,
        requireTransportSecurity=ssl_flag,
        requireAuthentication=need_login,)
    if ssl_flag:
        conn = reactor.connectSSL(host, port_, factory, cf)
    else:
        conn = reactor.connectTCP(host, port_, factory)
    return result


def send(filename, email, subject = 'datahaven_transport', reactor_stop=False, use_smtplib=False):
    transport_control.log('email', 'start sending to ' + email)
    smtp_info = [
            settings.getEmailAddress(),   #0
            settings.getSMTPHost(),       #1
            settings.getSMTPPort(),       #2
            settings.getSMTPUser(),       #3
            settings.getSMTPPass(),       #4
            settings.getSMTPSSL(),        #5
            settings.getSMTPNeedLogin()]  #6

    if smtp_info[0].strip() == '' or smtp_info[1].strip() == '':
        dhnio.Dprint(1, 'transport_email.send empty mailbox info. check your email options')
##        transport_control.msg('e', 'empty mailbox info. check your email options')
##        transport_control.connectionStatusCallback('email', email, None, 'empty mailbox info. check your email options')
        transport_control.sendStatusReport(email, filename,
            'failed', 'email',
            None, 'empty mailbox info. check your email options')
        return

##    print smtp_info
##    smtp_info = smtp3_info
    if not os.path.isfile(filename):
        dhnio.Dprint(1, 'transport_email.send wrong file name: %s' % filename)
##        transport_control.msg('e', 'wrong file name %s' % filename)
        return
    from_address = smtp_info[0]
    dhnio.Dprint(8, 'transport_email.send  from:%s to:%s' % (from_address, email))

    if use_smtplib:
        #send via smtplib
        try:
            send_smtplib(
                smtp_info[1],
                smtp_info[3],
                smtp_info[4],
                smtp_info[5],
                from_address,
                email,
                subject,
                filename)
        except:
            dhnio.Dprint(1, 'transport_email.send NETERROR sending email ' + dhnio.formatExceptionInfo())
##            transport_control.msg('e', 'Error sending email')
            transport_control.log('email', 'error sending email')
##            transport_control.connectionStatusCallback('email', email, None, 'error sending email')
            transport_control.sendStatusReport(email, filename,
                'failed', 'email',
                None, 'error sending email')
            return
##        if misc.transport_control_using():
        transport_control.sendStatusReport(email, filename,
            'finished', 'email')  # only one outstanding send per host
        transport_control.log('email', 'done')
        poll().send()
    else:
        #Send via Twisted
        def ok(x):
            dhnio.Dprint(14, 'transport_email.send.ok: sending done')
##            if misc.transport_control_using():
            # only one outstanding send per host
            transport_control.sendStatusReport(email, filename,
                'finished', 'email')
            transport_control.log('email', 'message successful sent')
            poll().send()
            if reactor_stop:
                reactor.stop()
        def fail(x):
##            if misc.transport_control_using():
            dhnio.Dprint(1, 'transport_email.send.fail NETERROR sending email \n' + str(x.getErrorMessage()))
##            transport_control.msg('e', 'error sending email: '  + str(x.getErrorMessage()) )
##            transport_control.connectionStatusCallback('email', smtp_info[0], x, 'error sending email')
            transport_control.sendStatusReport(email, filename,
                'failed', 'email',
                x, 'error sending email')
            transport_control.log('email', 'error sending email'  + str(x.getErrorMessage()))
            if reactor_stop:
                reactor.stop()

        if smtp_info[6]:
            #1th method
            res = sendTwistedMailAuth(
                smtp_info[1],
                smtp_info[2],
                smtp_info[3],
                smtp_info[4],
                smtp_info[5],
                smtp_info[6],
                from_address,
                email,
                subject,
                filename)
        else:
            #2th method:
            res = sendTwistedMail(
                smtp_info[1],
                smtp_info[0],
                email,
                subject,
                filename,
                smtp_info[2])
        res.addCallback(ok)
        res.addErrback(fail)
        return res


def send_public(filename, email, subject = 'datahaven_transport', reactor_stop=False, use_smtplib=False):
    smtp_info = [
            settings.getEmailAddress(),   #0
            settings.getSMTPHost(),       #1
            settings.getSMTPPort(),       #2
            settings.getSMTPUser(),       #3
            settings.getSMTPPass(),       #4
            settings.getSMTPSSL(),        #5
            settings.getSMTPNeedLogin()]  #6
    def ok(x):
        dhnio.Dprint(14, 'transport_email.send_public.ok: sending done')
##        if misc.transport_control_using():
        transport_control.sendStatusReport(email, filename,
            'finished', 'email')  # only one outstanding send per host
        transport_control.log('email', 'message successful sent')
        poll().send()
        if reactor_stop:
            reactor.stop()
    def fail(x):
        dhnio.Dprint(1, 'transport_email.send_public.fail NETERROR sending email \n' + str(x.getErrorMessage()))
##        transport_control.msg('e', 'error sending email: ' + str(x.getErrorMessage()))
##        transport_control.connectionStatusCallback('email', smtp_info[0], x, 'error sending email')
        transport_control.sendStatusReport(email, filename,
            'failed', 'email',
            x, 'error sending email')
        transport_control.log('email', 'NETERROR sending email')
        if reactor_stop:
            reactor.stop()

    res = sendTwistedMailAuth(
        smtp_info[1],
        smtp_info[2],
        smtp_info[3],
        smtp_info[4],
        smtp_info[5],
        smtp_info[6],
        smtp_info[0],
        email,
        subject,
        filename)
    res.addCallback(ok)
    res.addErrback(fail)
    return res


##############################################################
## Recevie POP ###################################################
##############################################################


def errorHandler(err):
    dhnio.Dprint(1, 'transport_email.errorHandler NETERROR \n' + str(err))
##    reactor.stop()
##    print err
##    raise err

class MyPOP3Client(POP3Client):
    linesize = MAX_LINE_LENGTH
    current_filename = ''
    def lineReceived(self, line):
        dhnio.Dprint(19, 'transport_email.lineReceived')
        if log_state:
            print '>>>:', line
##            print len(line)
        if len(line) <= self.linesize:
##            while line != '':
##                buf = line[:self.linesize]
##                POP3Client.lineReceived(self, buf+'\r\n')
##                line = line[self.linesize:]
##        else:
            POP3Client.lineReceived(self, line)

    def sendLine(self, line):
        dhnio.Dprint(19, 'transport_email.sendLine')
        if log_state:
            print '<<<:', line
        if len(line) <= self.linesize:
            POP3Client.sendLine(self, line)

    def connectionMade(self):
        dhnio.Dprint(14, 'transport_email.connectionMade')
        self.allowInsecureLogin = True
        self.startedTLS = True
        self.do_login()

##    def connectionLost(self, reason):
##        dhnio.Dprint(12, 'transport_email.connectionLost' + str(reason))

    def do_login(self):
        dhnio.Dprint(14, 'transport_email.do_login')
        self.login(self.factory.login,self.factory.password).addCallbacks(
            self.check_mailbox, self.err_login)

    def err_login(self, err):
        dhnio.Dprint(6, 'transport_email.err_login NETERROR ' + str(err))
##        transport_control.msg('e', 'error login on email server: ' + str(err))
##        transport_control.connectionStatusCallback('email', self.transport.getPeer(), err, 'error login on the email server')
        transport_control.receiveStatusReport('', 'failed',
            'email', self.transport.getPeer(),
             err, 'error login on the email server')
        self.prepare_loop('err_login')

    def check_mailbox(self, ignore):
        dhnio.Dprint(8, 'transport_email.check_mailbox')
        self.do_stat()

    def do_stat(self):
        self.stat().addCallbacks(
            self.set_stats, self.err_stat)

    def err_stat(self, err):
        dhnio.Dprint(6, 'transport_email.err_stat NETERROR ' + str(err))
##        transport_control.msg('e', 'can not retreive list of messages')
##        transport_control.connectionStatusCallback('email', self.transport.getPeer(), err, 'can not retreive list of messages')
        transport_control.receiveStatusReport('', 'failed',
            'email', self.transport.getPeer(),
             err, 'can not retreive list of messages')
        self.finish()

    def set_stats(self, stats):
        dhnio.Dprint(16, 'transport_email.set_stats')
        self.num_messages = stats[0]
        self.cur_message = 0
        transport_control.log('email', 'checking mailbox')
        if self.num_messages == 0:
            self.no_messages()
        else:
            self.retrieve(0).addCallbacks(
                self.do_retrieve_msg, self.err_retrieve_msg)

    def do_retrieve_msg(self, lines):
##        dhnio.Dprint(12, 'transport_email.do_retrieve_msg')
        msg_src = "\r\n".join(lines)
        msg = email.message_from_string(msg_src)
        self.process_msg(msg)

    def err_retrieve_msg(self, err):
        dhnio.Dprint(6, 'transport_email.err_retrieve_msg '+str(err))
##        print err
##        transport_control.msg('e', 'error retreive message')
##        transport_control.connectionStatusCallback('email', self.transport.getPeer(), err, 'error retreive message')
        transport_control.receiveStatusReport('', 'failed',
            'email', self.transport.getPeer(),
             err, 'error retreive message')
        self.finish()
##        raise err

    def process_msg(self, msg):
        dhnio.Dprint(12, 'transport_email.process_msg')
        take = False
        if msg['Subject'] == 'datahaven_transport':
            take = True
#            fout, self.current_filename = tempfile.mkstemp(".dhn-email-in")
            fout, self.current_filename = tmpfile.make("email-in")
            dhnio.Dprint(8, 'transport_email.message from: ' + str(msg['From']))
            transport_control.log('email', 'receiving file from ' + str(msg['From']))
            for part in msg.walk():
                if part.get_content_maintype() == 'multipart':
                    continue
                src = part.get_payload(decode=True)
                src = src[:-3]
                if src != None:
                    os.write(fout,src)
            os.close(fout)
##            transport_control.log('email', 'download... ')
            poll().receive()


        if take:
            self.delete(self.cur_message).addCallbacks(
                self.delete_msg, self.err_delete_msg)
        else:
            self.cur_message += 1
            if self.cur_message < self.num_messages:
                self.retrieve(self.cur_message).addCallbacks(
                    self.do_retrieve_msg, self.err_retrieve_msg)
            else:
                self.no_messages()

    def delete_msg(self, ignore):
        global ready_receiving
        dhnio.Dprint(14, 'transport_email.delete_msg')
        self.cur_message += 1
        if self.cur_message < self.num_messages:
            self.retrieve(self.cur_message).addCallbacks(
                self.do_retrieve_msg, self.err_retrieve_msg)
        else:
            self.no_messages()

        dhnio.Dprint(14, 'transport_email: received file ' + self.current_filename)
        if self.factory.receive_event is not None:
            self.factory.receive_event(self.current_filename)

        transport_control.log('email', 'receiving done. remove message from server')
##        if misc.transport_control_using() and ready_receiving:
        if ready_receiving:
            transport_control.receiveStatusReport(self.current_filename, "finished",
            'email', self.transport.getPeer())

    def err_delete_msg(self, err):
        dhnio.Dprint(6, 'transport_email.err_delete_msg NETERROR')
##        transport_control.msg('e', 'can not delete message from the server')
##        transport_control.connectionStatusCallback('email', self.transport.getPeer(), err, 'can not delete message from the server')
        transport_control.receiveStatusReport('', 'failed',
            'email', self.transport.getPeer(),
             err, 'can not delete message from the server')
        self.prepare_loop()
##        raise err

    def no_messages(self):
        global ready_receiving
        dhnio.Dprint(16, 'transport_email.no_messages')
        if not ready_receiving:
            dhnio.Dprint(8, 'transport_email.no_messages: mailbox is empty')
            transport_control.log('email', 'MAILBOX IS EMPTY. ready to receive new messages')
            ready_receiving = True
##        if self.factory.no_messages_event is not None:
##            self.factory.no_messages_event()
        self.finish()

    def finish(self):
        dhnio.Dprint(16, 'transport_email.finish')
        self.quit().addCallbacks(
            self.do_logout, self.err_logout)

    def do_logout(self, ignore):
        dhnio.Dprint(14, 'transport_email.do_logout')
###        timeout = misc.getEmailPollingTime()
##        timeout = pollingInterval()
##        transport_control.log('email', 'logout. polling time is: ' + str(timeout))
##        dhnio.Dprint(14, 'transport_email.do_logout. polling time is: ' + str(timeout))
##        timeout = self.factory.timeout
##        timeout = misc.ReceiveTimeOutEmail()
##        transport_control.log('email', 'current traffic: ' + str(poll().traffic()))
##        delaycall = reactor.callLater(timeout, self.factory.loop_callback)
        delaycall = self.prepare_loop()
        if self.factory.finish_event is not None:
            self.factory.finish_event(delaycall)
        self.transport.loseConnection()

    def prepare_loop(self, status=''):
        timeout = pollingInterval()
        if status == 'err_login':
            timeout *= 20
        dhnio.Dprint(10, 'transport_email.prepare_loop timeout=' + str(timeout))
        delaycall = reactor.callLater(timeout, self.factory.loop_callback)
        return delaycall

    def err_logout(self, err):
        dhnio.Dprint(6, 'transport_email.err_logout NETERROR' + str(err))
##        transport_control.msg('e', 'error logout from the server')
##        transport_control.connectionStatusCallback('email', self.transport.getPeer(), err, 'error logout from the server')
        transport_control.receiveStatusReport('', 'failed',
            'email', self.transport.getPeer(),
             err, 'error logout from the server')
        self.prepare_loop()
##        raise err

class PopChecker(ClientFactory):
    protocol = MyPOP3Client
    def __init__(self,
            login, password,
            timeout, loop_callback,
            receive_event=None,
            finish_event=None,
            no_messages_event=None):
        dhnio.Dprint(14, 'transport_email.PopChecker.__init__')
        self.login = login
        self.password = password
        self.timeout = timeout
        self.loop_callback = loop_callback
        self.receive_event = receive_event
        self.finish_event = finish_event

##    def clientConnectionLost(self, connector, reason):
##        dhnio.Dprint(12, 'transport_email.PopChecker.clientConnectionLost')
##        def reconnect():
##            connector.connect()
##
##    def clientConnectionFailed(self, connector, reason):
##        dhnio.Dprint(12, 'transport_email.PopChecker.clientConnectionFailed')

##def _loop():
##    dhnio.Dprint(12, 'transport_email._loop')
##    c = PopChecker(pop_login, pop_password, pop_timeout, _loop)
##    reactor.connectTCP(pop_server, pop_port, c)

class PopCheckerServer(Deferred):
    def __init__(self, server, port, login, password,
            ssl_flag, timeout=5,
            receive_event=None, no_messages_event=None):
        Deferred.__init__(self)
        dhnio.Dprint(16, 'transport_email.PopCheckerServer.__init__')
        self.server = server
        try:
            self.port = int(port)
        except:
            self.port = 110
        self.ssl_flag = ssl_flag
        self.addCallbacks(self.stop,self.fail)
        self.loopcall = None
        self.shutdown = False
        self.checker = PopChecker(login, password,
            timeout, self.run, receive_event,
            finish_event=self.finish_callback,
            no_messages_event=no_messages_event)

    def finish_callback(self, delaycall):
        dhnio.Dprint(14, 'transport_email.PopCheckerServer.finish_callback ' + str(delaycall))
        self.loopcall = delaycall

    def run(self):
        dhnio.Dprint(14, 'transport_email.PopCheckerServer.run')
        poll().update()
##        poll().clear()
        if self.shutdown:
            dhnio.Dprint(14, 'transport_email.PopCheckerServer.run - NEED SHUTDOWN')
            return
        if self.ssl_flag == 'True':
            self.runner = reactor.connectSSL(self.server, int(self.port), self.checker, ClientContextFactory())
            dhnio.Dprint(14, 'transport_email.connectSSL: %s: %s'% (self.server, str(self.port)))
        else:
            self.runner = reactor.connectTCP(self.server, int(self.port), self.checker)
            dhnio.Dprint(14, 'transport_email.connectTCP: %s: %s'% (self.server, str(self.port)))

    def stop(self, a):
        dhnio.Dprint(14, 'transport_email.PopCheckerServer.stop')
        self.stopListening()

    def fail(self, a):
        dhnio.Dprint(14, 'transport_email.PopCheckerServer.fail')
        self.stopListening()

    def stopListening(self):
        dhnio.Dprint(14, 'transport_email.PopCheckerServer.stopListening')
##        try:
##        print dir(self.checker.protocol) #.exit()
##        print self.checker.protocol.stop_me
##        print self.loopcall
        if self.loopcall is not None:
            try:
                self.loopcall.cancel()
            except:
                dhnio.Dprint(14, 'transport_email.PopCheckerServer.stopListening - loop call already canceled')
##        print self.runner
        self.runner.disconnect()

        return succeed(1)
##            self.checker.stopFactory()
##        except:
##            dhnio.Dprint(6, "errro. i can't disconnect")

#email_info = [email, pop3_server, port, login, password]
def receive(email_info, timeout=15, receive_event=None):
    transport_control.log('email', 'start receiving from '+str(email_info[0]))
    dhnio.Dprint(8, 'transport_email.receive. start looking at [%s:%s]  user: %s' % (email_info[1], email_info[2], email_info[3]))
##    print email_info
    server = PopCheckerServer(email_info[1], email_info[2], email_info[3], email_info[4], email_info[5], timeout, receive_event)
    server.loopcall = reactor.callLater(0, server.run)
    return server


##############################################################
### INIT #####################################################
##############################################################

def init():
    dhnio.Dprint(4, 'transport_email.init')
    global ready_receiving
    ready_receiving = False
    poll().init()

##    email_info = [
##            settings.getEmailAddress(),
##            settings.getPOPHost(),
##            settings.getPOPPort(),
##            settings.getPOPUser(),
##            settings.getPOPPass(),
##            settings.getPOPSSL()]
##
##    transport_control.log('email', 'cleaning email account...')
##    dhnio.Dprint(5, 'transport_email.init. start reading at [%s:%s]  user: %s' % (email_info[1], email_info[2], email_info[3]))
####    print email_info
##    mylistener = None
##    def no_messages():
##        transport_control.log('email','cleaning done')
##        mylistener.stopListening()
##    mylistener = PopCheckerServer(
##        email_info[1],
##        email_info[2],
##        email_info[3],
##        email_info[4],
##        email_info[5],
##        20,
##        no_messages_event=no_messages)
##    mylistener.loopcall = reactor.callLater(0, mylistener.run)


##############################################################
### private tests ############################################
##############################################################

##def my_tests():
####    _loop()
####    send('send-webdav.py',to_address)
##    send('misc.pyc',to_address)
####    reactor.run()
##
####my_tests()
##
##if __name__ == '__main__':
##    main()
##    reactor.run()
##

#-------------------------------------------------------------------------------

def mytest():
    dhnio.SetDebug(30)

##    filename = "identity.pyc"              # just some file to send

    pop_info = [
        settings.getEmailAddress(),
        settings.getPOPHost(),
        settings.getPOPPort(),
        settings.getPOPUser(),
        settings.getPOPPass(),
        settings.getPOPSSL()]

    smtp_info = [
        settings.getEmailAddress(),
        settings.getSMTPHost(),
        settings.getSMTPPort(),
        settings.getSMTPUser(),
        settings.getSMTPPass(),
        settings.getSMTPSSL()]



##    smtp_info = smtp2_info


    if len(sys.argv) == 1:
        print pop_info
        if not check_info(pop_info):
            print 'need POP info. see user configuration file'
            sys.exit()
        mylistener = None
        def receive_file(x):
            print 'receive 1 file.', x
##            mylistener.stopListening()
        mylistener = receive(pop_info, receive_event=receive_file)
##        reactor.run()
    elif len(sys.argv) == 3:
##        send(filename, pop_info[0])
        send(sys.argv[2], sys.argv[1], reactor_stop=True, use_smtplib=False)
    elif len(sys.argv) == 4 and sys.argv[1].lower() == '-p':
        send_public(sys.argv[3], sys.argv[2], reactor_stop=True, use_smtplib=False)
    else:
        print 'usage:'
        print 'to receive:   transport_email'
        print 'to send   :   transport_email [e-mail] [filename]'
        print 'to send from public datahaven server:'
        print '              transport_email -p [e-mail] [filename]'
        sys.exit(0)


if __name__ == "__main__":
    init()
    mytest()

##    keys = sys.modules.keys()
##    keys.sort()
##    for k in keys:
##        print str(k) + ' ' + str(sys.modules[k]) + '\n\n'

    reactor.run()





#!/usr/bin/python
#data_sender.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2013
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

import os
import sys
import time

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in data_sender.py')

import lib.dhnio as dhnio
import lib.misc as misc
import lib.packetid as packetid
import lib.contacts as contacts
import lib.settings as settings
import lib.diskspace as diskspace
import lib.nameurl as nameurl
import lib.transport_control as transport_control
import lib.automat as automat
import lib.automats as automats

import io_throttle
import backups
import fire_hire
import contact_status
import backup_monitor

_DataSender = None

#------------------------------------------------------------------------------ 

def A(event=None, arg=None):
    global _DataSender
    if _DataSender is None:
        _DataSender = DataSender('data_sender', 'READY', 6)
    if event is not None:
        _DataSender.automat(event, arg)
    return _DataSender


class DataSender(automat.Automat):
    timers = {'timer-1min':     (60,     ['READY']),
              'timer-1sec':     (5,      ['SENDING'])
              }
    statistic = {}

    def state_changed(self, oldstate, newstate):
        automats.set_global_state('DATASEND ' + newstate)

    def A(self, event, arg):
        #---READY---
        if self.state is 'READY':
            if event == 'new-data' or event == 'timer-1min' or event == 'restart' :
                self.state = 'SCAN_BLOCKS'
                self.doScanAndQueue(arg)
            elif event == 'init' :
                pass
        #---SCAN_BLOCKS---
        elif self.state is 'SCAN_BLOCKS':
            if event == 'scan-done' and self.isQueueEmpty(arg) :
                self.state = 'READY'
                self.doRemoveUnusedFiles(arg)
                backup_monitor.A('restart')
            elif event == 'scan-done' and not self.isQueueEmpty(arg) :
                self.state = 'SENDING'
        #---SENDING---
        elif self.state is 'SENDING':
            if event == 'restart' or ( ( event == 'block-acked' or event == 'block-failed' ) and self.isQueueEmpty(arg) ) :
                self.state = 'SCAN_BLOCKS'
                self.doScanAndQueue(arg)
            elif event == 'timer-1sec' :
                self.doPrintStats(arg)

    def isQueueEmpty(self, arg):
        if not arg:
            return io_throttle.IsSendingQueueEmpty()
        remoteID, packetID = arg
        return io_throttle.OkToSend(remoteID)
    
#    def isMoreDataToTransfer(self, arg):
#        for backupID in backups.GetBackupIds():
#            for blockNum in range(backups.GetKnownMaxBlockNum(backupID) + 1):
#                remoteArray = {'Data':   backups.GetRemoteDataArray(backupID, blockNum),
#                               'Parity': backups.GetRemoteParityArray(backupID, blockNum)}
#                localArray =  {'Data':   backups.GetLocalDataArray(backupID, blockNum),
#                               'Parity': backups.GetLocalParityArray(backupID, blockNum)}
#                for supplierNum in range(len(remoteArray)):
#                    for DP in ['Data', 'Parity']:
#                        if remoteArray[DP][supplierNum] != 1 and localArray[DP][supplierNum] == 1:
#                            return True
#        return False 
    
    def doScanAndQueue(self, arg):
        dhnio.Dprint(6, 'data_sender.doScanAndQueue')
        if '' not in contacts.getSupplierIDs():
            for backupID in misc.sorted_backup_ids(backups.local_files().keys()):
                packetsBySupplier = backups.ScanBlocksToSend(backupID)
                for supplierNum in packetsBySupplier.keys():
                    supplier_idurl = contacts.getSupplierID(supplierNum)
                    if not supplier_idurl:
                        dhnio.Dprint(2, 'data_sender.doScanAndQueue WARNING ?supplierNum? %s for %s' % (supplierNum, backupID))
                        continue
                    for packetID in packetsBySupplier[supplierNum]:
                        backupID_, blockNum, supplierNum_, dataORparity = packetid.BidBnSnDp(packetID)
                        if backupID_ != backupID:
                            dhnio.Dprint(2, 'data_sender.doScanAndQueue WARNING ?backupID? %s for %s' % (packetID, backupID))
                            continue
                        if supplierNum_ != supplierNum:
                            dhnio.Dprint(2, 'data_sender.doScanAndQueue WARNING ?supplierNum? %s for %s' % (packetID, backupID))
                            continue
                        if io_throttle.HasPacketInSendQueue(supplier_idurl, packetID):
                            continue
                        if not io_throttle.OkToSend(supplier_idurl):
                            continue
                        if len(transport_control.transfers_by_idurl(supplier_idurl)) > 0:
                            continue
                        filename = os.path.join(settings.getLocalBackupsDir(), packetID)
                        if not os.path.isfile(filename):
                            continue
                        io_throttle.QueueSendFile(
                            filename, 
                            packetID, 
                            supplier_idurl, 
                            misc.getLocalID(), 
                            self.packetAcked, 
                            self.packetFailed)
                        # dhnio.Dprint(6, '  %s for %s' % (packetID, backupID))
        self.automat('scan-done')
        
#        dhnio.Dprint(2, 'data_sender.doRemoveUnusedFiles [%s] [%s]' % (self.remoteData, self.remoteParity))
#        for supplierNum in range(self.supplierCount):
#            # data_filename = os.path.join(tmpfile.subdir('data-par'), self.BuildFileName(supplierNum, 'Data'))
#            # parity_filename = os.path.join(tmpfile.subdir('data-par'), self.BuildFileName(supplierNum, 'Parity'))
#            data_filename = os.path.join(settings.getLocalBackupsDir(), self.BuildFileName(supplierNum, 'Data'))
#            parity_filename = os.path.join(settings.getLocalBackupsDir(), self.BuildFileName(supplierNum, 'Parity'))
#            if os.path.isfile(data_filename) and self.remoteData[supplierNum] == 1:
#                try:
#                    os.remove(data_filename)
#                    #dhnio.Dprint(6, '    ' + os.path.basename(data_filename))
#                except:
#                    dhnio.DprintException()
#            if os.path.isfile(parity_filename) and self.remoteParity[supplierNum] == 1:
#                try:
#                    os.remove(parity_filename)
#                    #dhnio.Dprint(6, '    ' + os.path.basename(parity_filename))
#                except:
#                    dhnio.DprintException()

    def doPrintStats(self, arg):
        if dhnio.Debug(18):
            transfers = transport_control.current_transfers()
            bytes_stats = transport_control.current_bytes_transferred()
            s = ''
            for info in transfers:
                s += '%s ' % (diskspace.MakeStringFromBytes(bytes_stats[info.transfer_id]).replace(' ', '').replace('bytes', 'b'))
            dhnio.Dprint(0, 'transfers: ' + s[:120])

    def doRemoveUnusedFiles(self, arg):
        # we want to remove files for this block 
        # because we only need them during rebuilding
        if settings.getGeneralLocalBackups() is True:
            # if user set this in settings - he want to keep the local files
            return
        # ... user do not want to keep local backups
        if settings.getGeneralWaitSuppliers() is True:
            # but he want to be sure - all suppliers are green for long time
            if contact_status.hasOfflineSuppliers() or time.time() - fire_hire.GetLastFireTime() < 24*60*60:
                # some people are not there or we do not have stable team yet
                # do not remove the files because we need it to rebuild
                return 
        dhnio.Dprint(6, 'data_sender.doRemoveUnusedFiles')
        for backupID in misc.sorted_backup_ids(backups.local_files().keys()):
            packets = backups.ScanBlocksToRemove(backupID, settings.getGeneralWaitSuppliers())
            for packetID in packets:
                filename = os.path.join(settings.getLocalBackupsDir(), packetID)
                if os.path.isfile(filename):
                    try:
                        os.remove(filename)
                        dhnio.Dprint(6, '    ' + os.path.basename(filename))
                    except:
                        dhnio.DprintException()
        backups.ReadLocalFiles()
                         

    def packetAcked(self, packet, ownerID, packetID):
        backupID, blockNum, supplierNum, dataORparity = packetid.BidBnSnDp(packetID)
        backups.RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, True)
        if not self.statistic.has_key(ownerID):
            self.statistic[ownerID] = [0, 0]
        self.statistic[ownerID][0] += 1
        self.automat('block-acked', (ownerID, packetID))
    
    def packetFailed(self, remoteID, packetID, why):
        backupID, blockNum, supplierNum, dataORparity = packetid.BidBnSnDp(packetID)
        backups.RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, False)
        if not self.statistic.has_key(remoteID):
            self.statistic[remoteID] = [0, 0]
        self.statistic[remoteID][1] += 1
        self.automat('block-failed', (remoteID, packetID))


def statistic():
    global _DataSender
    if _DataSender is None:
        return {}
    return _DataSender.statistic
    
        
        
        
        
        
        
        
        




from struct import pack, unpack

from ncrypt0.ssl import SSLWantReadError, SSLWantWriteError, SSLWantError, SSLError

from nitro.async import AsyncOp

def sslAccept( sslConn, reactor, callback=None ) :
    return SSLInitiator( sslConn, SSLInitiator.ACCEPT, reactor, callback ).getOp()

def sslConnect( sslConn, reactor, callback=None ) :
    return SSLInitiator( sslConn, SSLInitiator.CONNECT, reactor, callback ).getOp()

class SSLInitiator( object ) :
    ACCEPT = 0
    CONNECT = 1

    WANT_NONE = 0
    WANT_READ = 1
    WANT_WRITE = 2

    def __init__( self, sslConn, initiateCode, reactor, callback=None ) :
        self.sslConn = sslConn
        self.reactor = reactor
        if initiateCode == self.ACCEPT :
            self.sslFunc = self.sslConn.accept
        else :
            assert initiateCode == self.CONNECT 
            self.sslFunc = self.sslConn.connect
        self.sock = self.sslConn.sock
        self.sock.setblocking( 0 )
        timerOp = self.reactor.callLater( 0, self._doAction )
        self.op = AsyncOp( callback, timerOp.cancel )
        self.want = self.WANT_NONE

    def getOp( self ) : return self.op

    def _removeRead( self ) :
        self.reactor.removeReadCallback( self.sock.fileno() )
        self.want = self.WANT_NONE

    def _removeWrite( self ) :
        self.reactor.removeWriteCallback( self.sock.fileno() )
        self.want = self.WANT_NONE

    def _doAction( self ) :
        try :
            self.sslFunc()
        except SSLWantReadError :
            if self.want != self.WANT_READ :
                if self.want == self.WANT_WRITE : self._removeWrite()
                self.reactor.addReadCallback( self.sock.fileno(), self._doAction )
                self.want = self.WANT_READ
                self.op.setCanceler( self._removeRead )
            return
        except SSLWantWriteError :
            if self.want != self.WANT_WRITE :
                if self.want == self.WANT_READ : self._removeRead()
                self.reactor.addWriteCallback( self.sock.fileno(), self._doAction )
                self.want = self.WANT_WRITE
                self.op.setCanceler( self._removeWrite )
            return
        except SSLError, e :
            if self.want == self.WANT_READ : self._removeRead()
            elif self.want == self.WANT_WRITE : self._removeWrite()
            self.op.notify( e )
            return
        except:
            raise
        if self.want == self.WANT_READ : self._removeRead()
        elif self.want == self.WANT_WRITE : self._removeWrite()
        self.op.notify( None )

class SSLWriter( object ) :
    def __init__( self, sslConn, pendingWrite, reactor, timeout=60 ) :
        self.sslConn = sslConn
        self.pendingWrite = pendingWrite
        self.reactor = reactor
        self.timeout = timeout
        self.sock = sslConn.sock
        self.maxWriteChunk = 32768
        self.writeCompleteCallback = None
        self.errorCallback = None
        self.timeoutCallback = None
        self.timeoutOp = self.reactor.callLater( self.timeout, self._onTimeout )
        self.timerOp = self.reactor.callLater( 0, self._onAction )
        self.hasRead = self.hasWrite = False

    def setWriteCompleteCallback( self, writeCompleteCallback ) :
        self.writeCompleteCallback = writeCompleteCallback

    def setErrorCallback( self, errorCallback ) :
        self.errorCallback = errorCallback

    def setTimeoutCallback( self, timeoutCallback ) :
        self.timeoutCallback = timeoutCallback

    def shutdown( self ) :
        if self.timerOp :
            self.timerOp.cancel()
            self.timerOp = None
        if self.timeoutOp :
            self.timeoutOp.cancel()
            self.timeoutOp = None
        if self.hasRead :
            assert not self.hasWrite
            self.reactor.removeReadCallback( self.sock.fileno() )
            self.hasRead = False
        elif self.hasWrite :
            self.reactor.removeWriteCallback( self.sock.fileno() )
            self.hasWrite = False

    def onWriteComplete( self ) :
        if not self.writeCompleteCallback :
            raise NotImplementedError
        self.writeCompleteCallback()

    def onError( self, sslErr ) :
        if self.errorCallback :
            raise NotImplementedError
        self.errorCallback( sslErr )

    def onTimeout( self ) :
        if not self.timeoutCallback :
            raise NotImplementedError
        self.timeoutCallback()

    def _onAction( self ) :
        self.timerOp = None
        self._doAction()

    def _doAction( self ) :
        try :
            while self.pendingWrite :
                written = self.sslConn.send( self.pendingWrite[:self.maxWriteChunk] )
                self.pendingWrite = self.pendingWrite[written:]
        except SSLWantReadError :
            if not self.hasRead :
                if self.hasWrite :
                    self.reactor.removeWriteCallback( self.sock.fileno() )
                    self.hasWrite = False
                self.reactor.addReadCallback( self.sock.fileno(), self._doAction )
                self.hasRead = True
            return
        except SSLWantWriteError :
            if not self.hasWrite :
                if self.hasRead :
                    self.reactor.removeReadCallback( self.sock.fileno() )
                    self.hasRead = False
                self.reactor.addWriteCallback( self.sock.fileno() )
                self.hasWrite = True
            return
        except SSLError, e :
            self.shutdown()
            self.onError( e )
            return
        self.shutdown()
        self.onWriteComplete()

    def _onTimeout( self ) :
        self.timeoutOp = None
        self.shutdown()
        self.onTimeout()

def sslAbort( sslConn ) :
    sslConn.sock.close()

def sslClose( sslConn, reactor, callback=None, timeout=30 ) :
    return SSLCloser( sslConn, reactor, callback, timeout ).getOp()

class SSLCloser( object ) :
    WANT_NONE = 0
    WANT_READ = 1
    WANT_WRITE = 2
    def __init__( self, sslConn, reactor, callback=None, timeout=30 ) :
        self.sslConn = sslConn
        self.reactor = reactor
        self.timeout = timeout
        self.sock = sslConn.sock
        self.want = self.WANT_NONE
        self.timerOp = self.reactor.callLater( 0, self._onAction )
        self.timeoutOp = self.reactor.callLater( self.timeout, self._onTimeout )
        self.op = AsyncOp( callback, self._close )

    def getOp( self ) : return self.op

    def _close( self ) :
        if self.timerOp : self.timerOp.cancel()
        if self.timeoutOp : self.timeoutOp.cancel()
        if self.want == self.WANT_READ : self._removeRead()
        elif self.want == self.WANT_WRITE : self._removeWrite()
        self.sock.close()

    def _closeNotify( self ) :
        self._close()
        self.op.notify()

    def _removeRead( self ) :
        assert self.want == self.WANT_READ
        self.reactor.removeReadCallback( self.sock.fileno() )
        self.want = self.WANT_NONE

    def _removeWrite( self ) :
        assert self.want == self.WANT_WRITE
        self.reactor.removeWriteCallback( self.sock.fileno() )
        self.want = self.WANT_NONE

    def _onAction( self ) :
        self.timerOp = None
        self._doAction()

    def _doAction( self ) :
        try :
            result = self.sslConn.shutdown()
            if result == 0 :
                result = self.sslConn.shutdown()
        except SSLWantReadError :
            if self.want != self.WANT_READ :
                if self.want == self.WANT_WRITE : self._removeWrite()
                self.reactor.addReadCallback( self.sock.fileno(), self._doAction )
                self.want = self.WANT_READ
            return
        except SSLWantWriteError :
            if self.want != self.WANT_WRITE :
                if self.want == self.WANT_READ : self._removeRead()
                self.reactor.addWriteCallback( self.sock.fileno(), self._doAction )
                self.want = self.WANT_WRITE
            return
        except SSLError, e :
            self._closeNotify()
            return
        self._closeNotify()

    def _onTimeout( self ) :
        self.timeoutOp = None
        self._closeNotify()

class SSLReactor( object ) :
    INITIAL = 0
    SUCCESS = 1
    WANT_READ = 2
    WANT_WRITE = 3
    def __init__( self, sslConn, reactor, readCallback, writeCallback ) :
        self.sslConn = sslConn
        self.reactor = reactor
        self.readCallback = readCallback
        self.writeCallback = writeCallback
        self.sock = sslConn.sock
        self.readEnabled = False
        self.writeEnabled = False
        self.readState = self.INITIAL
        self.writeState = self.INITIAL

        self.hasRead = self.hasWrite = False
        self.timerOp = None

    def getSSL( self ) : return self.sslConn

    def shutdown( self ) :
        self.readEnabled = self.writeEnabled = False
        self._updateEvents()

    def enableRead( self, enable=True ) :
        if enable != self.readEnabled :
            self.readEnabled = enable
            self._updateEvents()

    def isReadEnabled( self ) : return self.readEnabled

    def enableWrite( self, enable=True ) :
        if enable != self.writeEnabled :
            self.writeEnabled = enable
            self._updateEvents()

    def isWriteEnabled( self ) : return self.writeEnabled

    def read( self, bufSize ) :
        self.readState = self.SUCCESS
        try :
            data = self.sslConn.recv( bufSize )
        except SSLWantReadError :
            self.readState = self.WANT_READ
            self._updateEvents()
            raise
        except SSLWantWriteError :
            self.readState = self.WANT_WRITE
            self._updateEvents()
            raise
        return data

    def write( self, data ) :
        self.writeState = self.SUCCESS
        try :
            written = self.sslConn.send( data )
        except SSLWantReadError :
            self.writeState = self.WANT_READ
            self._updateEvents()
            raise
        except SSLWantWriteError :
            self.writeState = self.WANT_WRITE
            self._updateEvents()
            raise
        return written

    def _updateEvents( self ) :
        regRead = regWrite = regTimer = False
        if self.readEnabled :
            pending = self.sslConn.pending()
            if self.readState == self.INITIAL :
                regTimer = True
            elif self.readState == self.SUCCESS :
                if pending > 0 : regTimer = True
                else : regRead = True
            elif self.readState == self.WANT_READ :
                regRead = True
            else :
                assert self.readState == self.WANT_WRITE
                regWrite = True
        if self.writeEnabled :
            if self.writeState == self.INITIAL :
                regTimer = True
            elif self.writeState == self.SUCCESS :
                regWrite = True
            elif self.writeState == self.WANT_READ :
                regRead = True
            else :
                assert self.writeState == self.WANT_WRITE
                regWrite = True
        if regRead != self.hasRead :
            self.hasRead = regRead
            if regRead :
                self.reactor.addReadCallback( self.sock.fileno(), self._onRead )
            else :
                self.reactor.removeReadCallback( self.sock.fileno() )
        if regWrite != self.hasWrite :
            self.hasWrite = regWrite
            if regWrite :
                self.reactor.addWriteCallback( self.sock.fileno(), self._onWrite )
            else :
                self.reactor.removeWriteCallback( self.sock.fileno() )
        if regTimer and not self.timerOp :
            self.timerOp = self.reactor.callLater( 0, self._onTimer )
        elif (not regTimer) and self.timerOp :
            self.timerOp.cancel()
            self.timerOp = None

    def _processEvent( self ) :
        if self.readEnabled : self.readCallback()
        if self.writeEnabled : self.writeCallback()
        self._updateEvents()

    def _onRead( self ) : self._processEvent()
    def _onWrite( self ) : self._processEvent()
    def _onTimer( self ) :
        self.timerOp = None
        self._processEvent()

class SSLStream( object ) :
    def __init__( self, sslConn, reactor ) :
        self.sslConn = sslConn
        self.reactor = reactor
        self.closeCallback = None
        self.errorCallback = None
        self.inputCallback = None
        self.writeCompleteCallback = None
        self.sslReactor = SSLReactor( sslConn, reactor, self._onRead, self._onWrite )
        self.pendingWrite = ''
        self.maxReadSize = 0
        self.hasShutdown = False

    def getSSL( self ) : return self.sslConn

    def setCloseCallback( self, closeCallback ) :
        self.closeCallback = closeCallback

    def setErrorCallback( self, errorCallback ) :
        self.errorCallback = errorCallback

    def setInputCallback( self, inputCallback ) :
        self.inputCallback = inputCallback

    def setWriteCompleteCallback( self, writeCompleteCallback ) :
        self.writeCompleteCallback = writeCompleteCallback

    def shutdown( self ) :
        assert not self.hasShutdown
        self.hasShutdown = True
        self.sslReactor.shutdown()
        self.pendingWrite = ''

    def close( self, deferred=False, callback=None ) :
        assert not self.hasShutdown
        if not deferred :
            self.shutdown()
            sslAbort( self.sslConn )
            return
        pendingWrite = self.pendingWrite
        self.shutdown()
        sslConn, reactor = self.sslConn, self.reactor
        if not pendingWrite :
            return sslClose( sslConn, reactor, callback )
        def doClose() :
            closeOp = sslClose( sslConn, reactor, op.notify )
            op.setCanceler( closeOp.cancel )
        writer = SSLWriter( sslConn, pendingWrite, reactor )
        writer.setWriteCompleteCallback( doClose )
        writer.setErrorCallback( lambda err : doClose() )
        writer.setTimeoutCallback( doClose )
        op = AsyncOp( callback, writer.shutdown )
        return op

    def initiateRead( self, maxReadSize ) :
        self.maxReadSize = maxReadSize
        self.sslReactor.enableRead( True )

    def cancelRead( self ) :
        self.sslReactor.enableRead( False )

    def writeData( self, data ) :
        self.pendingWrite += data
        self.sslReactor.enableWrite( True )

    def getPendingWrite( self ) :
        return self.pendingWrite

    def _notifyClose( self ) :
        self.sslReactor.enableRead( False )
        if self.closeCallback :
            self.closeCallback()
        else :
            self.close( deferred=True )

    def _notifyError( self, err ) :
        if self.errorCallback :
            self.errorCallback( err )
        else :
            self.close()

    def _notifyInput( self, data ) :
        if self.inputCallback :
            self.inputCallback( data )

    def _notifyWriteComplete( self ) :
        if self.writeCompleteCallback :
            self.writeCompleteCallback()

    def _onRead( self ) :
        try :
            data = self.sslReactor.read( self.maxReadSize )
            if not data :
                self._notifyClose()
                return
            self._notifyInput( data )
        except SSLWantError :
            return
        except SSLError, e :
            self._notifyError( e )

    def _onWrite( self ) :
        try :
            while self.pendingWrite :
                written = self.sslReactor.write( self.pendingWrite )
                self.pendingWrite = self.pendingWrite[written:]
            self.sslReactor.enableWrite( False )
            self._notifyWriteComplete()
        except SSLWantError :
            return
        except SSLError, e :
            self._notifyError( e )

class SSLMessageStream( object ) :
    def __init__( self, sslConn, reactor ) :
        self.stream = SSLStream( sslConn, reactor )
        self.readEnabled = False
        self.maxMsgLength = 65536
        self.readingLength = True
        self.msgLength = 0
        self.buffer = ''
        self.stream.setInputCallback( self._onInput )

        self.getSSL = self.stream.getSSL
        self.shutdown = self.stream.shutdown
        self.close = self.stream.close
        self.getPendingWrite = self.stream.getPendingWrite

        self.setCloseCallback = self.stream.setCloseCallback
        self.setErrorCallback = self.stream.setErrorCallback
        self.setWriteCompleteCallback = self.stream.setWriteCompleteCallback

        self.invalidMessageCallback = None
        self.inputCallback = None

    def setInvalidMessageCallback( self, invalidMessageCallback ) :
        self.invalidMessageCallback = invalidMessageCallback
    def setInputCallback( self, inputCallback ) :
        self.inputCallback = inputCallback

    def enableRead( self, enable ) :
        if enable != self.readEnabled :
            self.readEnabled = enable
            if enable :
                if self.readingLength :
                    assert len(self.buffer) < 4
                    self.stream.initiateRead( 4-len(self.buffer) )
                else :
                    assert len(self.buffer) < self.msgLength
                    self.stream.initiateRead( self.msgLength - len(self.buffer) )
            else :
                self.stream.cancelRead()

    def isReadEnabled( self ) : return self.readEnabled

    def sendMessage( self, data ) :
        msg = pack( '<i%ds' % len(data), len(data), data )
        self.stream.writeData( msg )

    def _onInput( self, data ) :
        assert self.readEnabled
        if self.readingLength :
            assert len(data) + len(self.buffer) <= 4
            self.buffer += data
            if len(self.buffer) == 4 :
                self.msgLength = unpack( '<i', self.buffer )[0]
                self.readingLength = False
                if (self.msgLength <= 0) or (self.msgLength > self.maxMsgLength) :
                    if self.invalidMessageCallback :
                        self.invalidMessageCallback()
                    else :
                        self.close()
                    return
                self.buffer = ''
                self.stream.initiateRead( self.msgLength )
            else :
                self.stream.initiateRead( 4-len(self.buffer) )
        else :
            assert len(data) + len(self.buffer) <= self.msgLength
            self.buffer += data
            if len(self.buffer) == self.msgLength :
                data = self.buffer
                self.buffer = ''
                self.msgLength = 0
                self.readingLength = True
                self.stream.initiateRead( 4 )
                if self.inputCallback :
                    self.inputCallback( data )
            else :
                self.stream.initiateRead( self.msgLength - len(self.buffer) )



'''
pywin32utils
PyWin32 Based Replacement for the C based CreateProcess Wrapper
'''
import sys

if sys.platform == 'win32' :
    import ctypes

    class win32_STARTUPINFO(ctypes.Structure):
        _fields_ = [('cb',ctypes.c_ulong),
                    ('lpReserved',ctypes.c_char_p),
                    ('lpDesktop',ctypes.c_char_p),
                    ('lpTitle',ctypes.c_char_p),
                    ('dwX',ctypes.c_ulong),
                    ('dwY',ctypes.c_ulong),
                    ('dwXSize',ctypes.c_ulong),
                    ('dwYSize',ctypes.c_ulong),
                    ('dwXCountChars',ctypes.c_ulong),
                    ('dwYCountChars',ctypes.c_ulong),
                    ('dwFillAttribute',ctypes.c_ulong),
                    ('dwFlags',ctypes.c_ulong),
                    ('wShowWindow',ctypes.c_ulong),
                    ('cbReserved2',ctypes.c_ulong),
                    ('lpReserved2',ctypes.c_ulong),
                    ('hStdInput',ctypes.c_void_p),
                    ('hStdOutput',ctypes.c_void_p),
                    ('hStdError',ctypes.c_void_p),
        ]

    class win32_PROCESS_INFORMATION(ctypes.Structure):
        _fields_ = [('hProcess',ctypes.c_void_p),
                    ('hThread',ctypes.c_void_p),
                    ('dwProcessId',ctypes.c_ulong),
                    ('dwThreadId',ctypes.c_ulong)
        ]

    def spawnProcess( appName, argv, env, startingDir, inheritHandles ) :
        cmdLine = []
        for arg in argv :
            if arg.find(' ') >= 0 :
                arg = '"%s"' % arg
            cmdLine.append( arg )
        cmdLine = ' '.join( cmdLine )
        environ = []
        for k,v in env.items() :
            environ.append( '%s=%s' % (k,v) )
        environ.append( '\x00' )
        environ = '\x00'.join( environ )
        startupInfo = win32_STARTUPINFO()
        processInfo = win32_PROCESS_INFORMATION()
        retval = ctypes.windll.kernel32.CreateProcessA(ctypes.c_char_p(appName),ctypes.c_char_p(cmdLine),None,None,inheritHandles,0,
        ctypes.c_char_p(environ),ctypes.c_char_p(startingDir),ctypes.pointer(startupInfo),ctypes.pointer(processInfo))
        ctypes.windll.kernel32.CloseHandle(processInfo.hProcess)
        ctypes.windll.kernel32.CloseHandle(processInfo.hThread)
        return retval
else :
    import subprocess

    def spawnProcess( appName, argv, env, startingDir, inheritHandles ) :
        p = subprocess.Popen( args=argv, executable=appName,
                close_fds=True, cwd=startingDir, env=env )
        return p.pid

__all__ = ['spawnProcess']



#!/usr/bin/python
#backup_db.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2013
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#1)  This tracks backups in metadata/backup_info.xml.
#    It saves the directories to be backed up,
#    the schedule for being backed up if there is one,
#    and then the backups,
#    so when our suppliers say they have files for a particular backup (F20110718100000AM),
#    we know what directory that corresponds to.
#    It should also send the backup_info.xml to the suppliers so that data is backed up.
#    TODO it should encrypt the file and send it out.
#
#2)  Tracks backups in process so we don't try to fix a backup that isn't finished,
#    also so we don't delete a backup in process.



import os
import sys
import time
import locale
import xml.dom.minidom

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor backup_db.py')

import lib.dhnio as dhnio
import lib.misc as misc
import lib.nameurl as nameurl
import lib.settings as settings
import lib.dirsize as dirsize
import lib.commands as commands
import lib.transport_control as transport_control
import lib.dhnpacket as dhnpacket
import lib.contacts as contacts
import lib.dhncrypto as dhncrypto
import lib.schedule as schedule

import dhnblock
import backup_db_keeper

#------------------------------------------------------------------------------ 

InitDone = False
_DB = None

#------------------------------------------------------------------------------ 

class _BackupRun:
    def __init__(self, backupID, backupSize=0, backupStatus="", backupStart="", backupFinish=""):
        self.backupID = backupID
        self.backupSize = backupSize
        self.backupStatus = backupStatus
        self.backupStart = backupStart
        self.backupFinish = backupFinish
        if str(self.backupStart) == "":
            self.backupStart = str(time.time())

    def SetStatus(self, backupStatus, backupFinish):
        self.backupStatus = backupStatus
        self.backupFinish = backupFinish

    def GetRunInfo(self):
        return self.backupID, self.backupSize, self.backupStatus, self.backupStart, self.backupFinish

    def Show(self):
        return "  " + self.backupID + "-" + self.backupStatus + "-" + self.backupSize + "-" + self.backupStart + "-" + self.backupFinish

    def ToXML(self):
        runXML =  "        <backupRun>\n"
        runXML += "          <backupId>" + self.backupID + "</backupId>\n"
        runXML += "          <backupSize>" + str(self.backupSize) + "</backupSize>\n"
        runXML += "          <backupStatus>" + self.backupStatus + "</backupStatus>\n"
        runXML += "          <backupStart>" + str(self.backupStart) + "</backupStart>\n"
        runXML += "          <backupFinish>" + str(self.backupFinish) + "</backupFinish>\n"
        runXML += "        </backupRun>\n"
        return runXML


class _BackupSchedule:
    def __init__(self, sched):
        self.schedule = sched

    def SetSchedule(self, sched):
        self.schedule = sched

    def Show(self):
        return self.schedule.description()

    def ToXML(self):
        scheduleXML =  "      <backupSchedule>\n"
        scheduleXML += "        <scheduleType>" + self.schedule.type + "</scheduleType>\n"
        scheduleXML += "        <scheduleTime>" + self.schedule.daytime + "</scheduleTime>\n"
        scheduleXML += "        <scheduleInterval>" + self.schedule.interval + "</scheduleInterval>\n"
        scheduleXML += "        <intervalDetails>" + self.schedule.details + "</intervalDetails>\n"
        scheduleXML += "      </backupSchedule>\n"
        return scheduleXML


class _BackupDirectory:
    def __init__(self, dirName, backupSubdirectories=True):
        self.dirName = dirName
        self.backupSubdirectories = backupSubdirectories # do backup of subdirectories
        self.schedule = None 
        self.backupRuns = [] # an array of _BackupRun in order of time, oldest to newest
        self.runningBackup = False # True if some backup of this folders is running 

    def AddBackupRun(self, backupID, backupSize, backupStatus, backupStart, backupFinish):
        self.backupRuns.append(_BackupRun(backupID, backupSize, backupStatus, backupStart, backupFinish))
        if backupStatus in ['in process', 'sending']:
            self.runningBackup = True

    def DeleteDirBackup(self, backupID):
        delBackupRun = []
        for backupRun in self.backupRuns:
            if backupRun.backupID == backupID:
                delBackupRun.append(backupRun)
        for backupRun in delBackupRun:
            self.backupRuns.remove(backupRun)
        del delBackupRun

    def GetLastRunInfo(self):
        if len(self.backupRuns)>0:
            return self.backupRuns[-1].GetRunInfo()
        else:
            return "", "0.0", "(none)", str(time.time()), ""

    def GetDirBackupIds(self):
        backupIds = []
        for backupRun in self.backupRuns:
            backupIds.append(str(backupRun.backupID))
        return backupIds

    def GetDirectorySubfoldersInclude(self):
        return self.backupSubdirectories

    def SetDirectorySubfoldersInclude(self, backupSubdirectories):
        if backupSubdirectories == False:
            self.backupSubdirectories = False
        else:
            self.backupSubdirectories = True

    def IsBackupRunning(self):
        return self.runningBackup

    def SetBackupStatus(self, backupID, backupStatus, backupFinish):
        for backupRun in self.backupRuns:  #TODO, do we need backupID at all?  Just set backupRuns[-1] ?
            if backupRun.backupID == backupID:
                dhnio.Dprint(6, 'backup_db.SetBackupStatus %s %s %s' % (backupID, backupStatus, str(backupFinish)))
                backupRun.SetStatus(backupStatus, backupFinish)
                self.runningBackup = False
                return

    def GetDirectoryInfo(self):
        if len(self.backupRuns) > 0:
            mostRecentBackupId = ""
            totalSize = 0.0
            mostRecentStatus = ""
            for backupRun in self.backupRuns:
                backupID, backupSize, backupStatus, backupStart, backupFinish = backupRun.GetRunInfo()
                totalSize += int(backupSize)
                mostRecentBackupId = backupID
                mostRecentStatus = backupStatus
            return mostRecentBackupId, totalSize, mostRecentStatus
        else:
            return '', 0, ''

    def SetDirSchedule(self, schedule):
        if self.schedule == None:
            self.schedule = _BackupSchedule(schedule)
        else:
            self.schedule.SetSchedule(schedule)

    def Show(self):
        dirShow =  self.dirName + "\n"
        if self.schedule:
            dirShow = dirShow + self.schedule.Show() + "\n"
        for backupRun in self.backupRuns:
            dirShow = dirShow + backupRun.Show() + "\n"
        return dirShow

    def ToXML(self):
        dirXML =  "    <backupDirectory>\n"
        dirXML += "      <directoryName>" + self.dirName + "</directoryName>\n"
        dirXML += "      <backupSubdirectories>" + str(self.backupSubdirectories) + "</backupSubdirectories>\n"
        if self.schedule:
            dirXML += self.schedule.ToXML()
        if len(self.backupRuns) > 0:
            dirXML += "      <directoryBackups>\n"
            for backupRun in self.backupRuns:
                dirXML += backupRun.ToXML()
            dirXML += "      </directoryBackups>\n"
        dirXML += "    </backupDirectory>\n"
        return dirXML


class _BackupDB:
    def __init__(self):
        self.backupDirs = {}
        self.deletedBackups = []
        self.revisionNumber = 0
        self.currentlyRunningBackups = {}
        self._xmlFile = settings.BackupInfoFileFullPath()
        self._maxDeleted = settings.MaxDeletedBackupIDsToKeep()
        self._loading = False

    def BackupDirs(self):
        return self.backupDirs

    def AskSizeForAllDirs(self):
        for backupDir in self.backupDirs.keys():
            dirsize.ask(backupDir)

    def _GetText(self, nodelist): # get text from DOM elements
        rc = ""
        for node in nodelist:
            if node.hasChildNodes():
                rc = rc + self._GetText(node.childNodes)
            elif node.nodeType == node.TEXT_NODE: # TEXT_NODE is node type 3
                rc = rc + node.data.strip()
        return rc

    def InboxBackupInfoPacket(self, packet):
        if packet.PacketID == settings.BackupInfoEncryptedFileName():
            try:
                block = dhnblock.Unserialize(packet.Payload)
                SessionKey = dhncrypto.DecryptLocalPK(block.EncryptedSessionKey)
                paddeddata = dhncrypto.DecryptWithSessionKey(SessionKey, block.EncryptedData)
                xmlSrc = paddeddata[:int(block.Length)]
                xmlDom = xml.dom.minidom.parseString(xmlSrc)
                rootNode = xmlDom.documentElement
            except:
                dhnio.Dprint(4, 'backup_db.InboxBackupInfoPacket ERROR reading xml source from %s' % packet.RemoteID)
                return
        else:
            try:
                xmlSrc = packet.Payload
                xmlDom = xml.dom.minidom.parseString(xmlSrc)
                rootNode = xmlDom.documentElement
            except:
                dhnio.Dprint(4, 'backup_db.InboxBackupInfoPacket ERROR reading xml source from %s' % packet.RemoteID)
                return
        revisionNumber = misc.ToInt(self._GetText(rootNode.getElementsByTagName('revisionNumber')), 0)
        if self.revisionNumber < revisionNumber:
            self.LoadXML(rootNode)
            self.Write()
        backup_db_keeper.A('incoming-db-info', packet)

    def ToXML(self):
        self.CommandLog("ToXML")
        dbXML = ''
        dbXML += '<?xml version="1.0" encoding="%s" ?>\n' % locale.getpreferredencoding()
        dbXML += '<backupData>\n'
        dbXML += '  <revisionNumber>%d</revisionNumber>\n' % self.revisionNumber 
        dbXML += '  <backupDirectories>\n'
        for dirName in self.backupDirs:
            dbXML += self.backupDirs[dirName].ToXML()
        dbXML += '  </backupDirectories>\n'
        dbXML += '  <deletedBackups>\n'
        for backupID in self.deletedBackups[0-self._maxDeleted:]:
            dbXML += '    <deletedBackup>\n'
            dbXML += '      <deletedBackupId>' + backupID + '</deletedBackupId>\n'
            dbXML += '    </deletedBackup>\n'
        dbXML += '  </deletedBackups>\n'
        dbXML += '</backupData>\n'
        return dbXML

    def LoadXML(self, rootNode):
        self.revisionNumber = misc.ToInt(self._GetText(rootNode.getElementsByTagName('revisionNumber')), 0)
        backupDirs = rootNode.getElementsByTagName("backupDirectory")
        for backupDir in backupDirs: # DOM Elements
            dirName = self._GetText(backupDir.getElementsByTagName("directoryName"))
            backupSubdirectories = 'true' == str(self._GetText(backupDir.getElementsByTagName("backupSubdirectories"))).lower().strip()
            self.AddDirectory(dirName, backupSubdirectories)
            for sched in backupDir.getElementsByTagName("backupSchedule"):
                self.SetDirSchedule(
                    dirName, 
                    schedule.Schedule( 
                        typ = str(self._GetText(sched.getElementsByTagName("scheduleType"))), 
                        daytime = str(self._GetText(sched.getElementsByTagName("scheduleTime"))), 
                        interval = str(self._GetText(sched.getElementsByTagName("scheduleInterval"))), 
                        details = str(self._GetText(sched.getElementsByTagName("intervalDetails")))),)
            for backupRun in backupDir.getElementsByTagName("backupRun"):
                self.AddDirBackup(
                    dirName, 
                    str(self._GetText(backupRun.getElementsByTagName("backupId"))), 
                    # when we loading backup_db from file or from remote supplier
                    # we do not care about backups statuses, only need to know backupID
                    # str(self._GetText(backupRun.getElementsByTagName("backupStatus"))),
                    'done', 
                    misc.ToInt(self._GetText(backupRun.getElementsByTagName("backupSize"))), 
                    misc.ToFloat(self._GetText(backupRun.getElementsByTagName("backupStart"))), 
                    misc.ToFloat(self._GetText(backupRun.getElementsByTagName("backupFinish"))),)
        deletedBackups = rootNode.getElementsByTagName("deletedBackup")
        for deletedBackup in deletedBackups:
            backupID = str(self._GetText(deletedBackup.getElementsByTagName("deletedBackupId")))
            self.deletedBackups.append(backupID)
    
    def Clear(self):
        self.backupDirs.clear()
        self.deletedBackups = []
        self.revisionNumber = 0

    def Write(self):
        try:
            src = self.ToXML()
            f = open(self._xmlFile+"new.xml", "w")
            f.write(src)
            f.close()
            try:
                xmlDOM = xml.dom.minidom.parse(self._xmlFile+"new.xml")
                # make sure we are saving something valid
            except:
                dhnio.Dprint(2, "backup_db.Write ERROR new file is not a valid XML file")
                dhnio.DprintException()
                os.remove(self._xmlFile+'new.xml')
                return
            if dhnio.Windows() and os.path.exists(self._xmlFile):
                os.remove(self._xmlFile)
            os.rename(self._xmlFile+"new.xml", self._xmlFile)
        except:
            dhnio.Dprint(2, "backup_db.Write ERROR unable to save new XML file")
            dhnio.DprintException()
        
    def Save(self):
        if self._loading:
            return
        self.revisionNumber += 1
        self.Write()
        backup_db_keeper.A('restart')

    def Load(self):
        if os.path.exists(self._xmlFile):
            self._loading = True
            try:
                xmlDOM = xml.dom.minidom.parse(self._xmlFile)
            except:
                dhnio.Dprint(2, "backup_db.Load ERROR not a valid XML file")
                return
            rootNode = xmlDOM.documentElement
            self.Clear()
            self.LoadXML(rootNode)
            self._loading = False

    # used during developement to debug some stuff
    def CommandLog(self, command):
        dhnio.Dprint(6, 'backup_db: ' + str(command))

    #--- begin of signing/encryption section, not yet active ---
    
    def SaveAddon(self):
        SessionKey = dhncrypto.NewSessionKey()
        # SessionKeyType = dhncrypto.SessionKeyType()
        # LocalID = misc.getLocalIdentity().getIDURL()
        Data = dhnio.ReadBinaryFile(self._xmlFile)
        DataLonger = misc.RoundupString(Data, 24)
        self.EncryptedData = dhncrypto.EncryptWithSessionKey(SessionKey, DataLonger)
        self.Signature = self.GenerateSignature()  # usually just done at packet creation

    def SessionKey(self):
        return dhncrypto.DecryptLocalPK(self.EncryptedSessionKey)

    def GenerateHashBase(self):
        sep = "::::"
        # PREPRO needs to have all fields and separator
        StringToHash = self.CreatorID + sep + self.SessionKeyType + sep + self.EncryptedSessionKey + sep + self.Length + sep + self.EncryptedData
        return StringToHash

    def GenerateHash(self):
        return dhncrypto.Hash(self.GenerateHashBase())

    def GenerateSignature(self):
        return dhncrypto.Sign(self.GenerateHash())
    
    #--- end of signing/encryption section, not yet active ---

    # dirName should be unicode!!!
    def AddDirectory(self, dirName, recursive=True):
        if not self.backupDirs.has_key(dirName):
            self.CommandLog("AddDirectory made new entry [%s]" % dirName)
            self.backupDirs[dirName] = _BackupDirectory(dirName, recursive)
        else:
            self.CommandLog("AddDirectory already exist {%s}" % dirName)

    def DeleteDirectory(self, dirName):
        self.CommandLog("DelDirectory " + dirName)
        for dbDir in self.backupDirs.keys():
            if os.path.abspath(dirName) == os.path.abspath(dbDir):
                for backupRun in self.backupDirs[dbDir].backupRuns:
                    if backupRun.backupID not in self.deletedBackups:
                        self.deletedBackups.append(backupRun.backupID)
                del self.backupDirs[dbDir]

    def CheckDirectory(self, dirName):
        if self.backupDirs.has_key(dirName):
            return True
        else:
            return False

    def GetBackupDirectories(self):
        backupDirectories = self.backupDirs.keys()
        # backupDirectories.sort()
        return backupDirectories

    def GetDirectoryInfo(self, dirName):
        if self.backupDirs.has_key(dirName):
            return self.backupDirs[dirName].GetDirectoryInfo()
        return '', 0, ''

    def AddDirBackup(self, dirName, backupID, backupStatus="", backupSize=0, backupStart=0, backupFinish=0, ):
        self.CommandLog("AddDirBackup [%s] %s %s %d %d %d" % (dirName, backupID, backupStatus, backupSize, backupStart, backupFinish,))
        if not self.backupDirs.has_key(dirName):
            self.backupDirs[dirName] = _BackupDirectory(dirName)
        self.backupDirs[dirName].AddBackupRun(backupID, backupSize, backupStatus, backupStart, backupFinish)

    def DeleteDirBackup(self, backupID):
        self.CommandLog("DeleteDirBackup " + backupID)
        dirName = self.GetDirectoryFromBackupId(backupID)
        if self.backupDirs.has_key(dirName):
            self.backupDirs[dirName].DeleteDirBackup(backupID)
        if backupID not in self.deletedBackups:
            self.deletedBackups.append(backupID)

    def IsDeleted(self, backupID):
        if backupID in self.deletedBackups:
            return True
        else:
            return False

    def GetDirLastBackup(self, dirName):
        if self.backupDirs.has_key(dirName):
            if len(self.backupDirs[dirName].backupRuns) > 0:
                lastBackup = self.backupDirs[dirName].backupRuns[-1]
                return lastBackup.backupID, lastBackup.backupSize, lastBackup.backupStatus, lastBackup.backupStart, lastBackup.backupFinish
        return "", 0, "", 0, 0

    def GetDirBackupIds(self, dirName):
        if self.backupDirs.has_key(dirName):
            return self.backupDirs[dirName].GetDirBackupIds()
        else:
            return []

    def SetBackupStatus(self, dirName, backupID, backupStatus, backupFinish):
        self.CommandLog("SetBackupStatus " + dirName + " " + backupID + " " + backupStatus + " " + backupFinish)
        if self.backupDirs.has_key(dirName):
            self.backupDirs[dirName].SetBackupStatus(backupID, backupStatus, backupFinish)

    def AbortDirectoryBackup(self, dirName):
        if self.backupDirs.has_key(dirName):
            self.backupDirs[dirName].runningBackup = False

    def GetLastRunInfo(self, dirName):
        return self.backupDirs[dirName].GetLastRunInfo()

    def IsBackupRunning(self, dirName):
        if self.backupDirs.has_key(dirName):
            return self.backupDirs[dirName].IsBackupRunning()
        return False

    def SetDirSchedule(self, dirName, sched):
        if self.backupDirs.has_key(dirName):
            self.CommandLog("SetDirSchedule %s %s %s %s %s" % (dirName, sched.type, sched.daytime, sched.interval, sched.details))
            self.backupDirs[dirName].SetDirSchedule(sched)

    def GetDirSchedule(self, dirName):
        if self.backupDirs.has_key(dirName):
            if self.backupDirs[dirName].schedule != None:
                return self.backupDirs[dirName].schedule.schedule
        return None

    def GetDirectorySubfoldersInclude(self, dirName):
        if self.backupDirs.has_key(dirName):
            return self.backupDirs[dirName].GetDirectorySubfoldersInclude()
        return True

    def SetDirectorySubfoldersInclude(self, dirName, backupSubfolders):
        if self.backupDirs.has_key(dirName):
            self.backupDirs[dirName].SetDirectorySubfoldersInclude(backupSubfolders)

    def GetDirectoryFromBackupId(self, backupID):
        for dirName in self.backupDirs.keys():
            for backupRun in self.backupDirs[dirName].backupRuns:
                if backupRun.backupID == backupID:
                    return dirName
        return ''

    def GetBackupIdRunInfo(self, backupID):
        for dirName in self.backupDirs.keys():
            for backupRun in self.backupDirs[dirName].backupRuns:
                if backupRun.backupID == backupID:
                    return backupRun.GetRunInfo()
        return None
    
    def GetBackupsByDateTime(self, reverse=False):
        d = {}
        for dirName in self.backupDirs.keys():
            for backupRun in self.backupDirs[dirName].backupRuns:
                d[backupRun.backupID] = (backupRun.backupID, dirName, backupRun.backupSize, backupRun.backupStatus)
        l = []
        for backupID in misc.sorted_backup_ids(d.keys(), reverse):
            l.append(d[backupID])
        d.clear()
        return l
    
    def GetBackupsByFolder(self):
        result = []
        for dirName in self.backupDirs.keys():
            dirBackupsSize = 0
            backupRuns = []
            for backupRun in self.backupDirs[dirName].backupRuns:
                backupRuns.append(backupRun)
                dirBackupsSize += backupRun.backupSize
            recentBackupID = '' if len(backupRuns) == 0 else backupRuns[0].backupID 
            result.append((dirName, backupRuns, dirBackupsSize, self.backupDirs[dirName].runningBackup, recentBackupID))
        return result 

    def Show(self):
        dbList = ""
        for dirName in self.backupDirs:
            dbList = dbList + self.backupDirs[dirName].Show()
        return dbList

    def GetBackupIds(self, full_info_please=False):
        backupIds = []
        for dirName in self.backupDirs:
            for backupRun in self.backupDirs[dirName].backupRuns:
                if full_info_please:
                    backupIds.append((str(backupRun.backupID), dirName, backupRun.backupSize, backupRun.backupStatus))
                else:
                    backupIds.append(str(backupRun.backupID))
        return misc.sorted_backup_ids(backupIds)

    def GetTotalBackupsSize(self):
        backupsSize = 0
        for dirName in self.backupDirs:
            for backupRun in self.backupDirs[dirName].backupRuns:
                backupsSize += int(backupRun.backupSize)
        return backupsSize

    # the backup objects are backup.backup's
    def AddRunningBackupObject(self, backupID, backupObject):
        if not self.currentlyRunningBackups.has_key(backupID):
            self.currentlyRunningBackups[backupID] = backupObject
            
    def GetRunningBackupObject(self, backupID):
        return self.currentlyRunningBackups.get(backupID, None)

    def RemoveRunningBackupObject(self, backupID):
        if self.currentlyRunningBackups.has_key(backupID):
            del self.currentlyRunningBackups[backupID]

    def AbortRunningBackup(self, backupID):
        if self.currentlyRunningBackups.has_key(backupID):
            self.currentlyRunningBackups[backupID].abort()
        
    def RemoveAllRunningBackupObjects(self):
        self.currentlyRunningBackups.clear()

    def ShowRunningBackups(self):
        return self.currentlyRunningBackups.keys()

    def HasRunningBackup(self):
        if len(self.currentlyRunningBackups.keys()) > 0:
            return True
        return False
    
#------------------------------------------------------------------------------ 

def BackupDirs():
    return db().BackupDirs()

def ToXML():
    return db().ToXML()

def Save():
    return db().Save()

def InboxBackupInfoPacket(packet):
    return db().InboxBackupInfoPacket(packet)

def AddDirectory(dirName, recursive=True):
    return db().AddDirectory(dirName, recursive)
    
def DeleteDirectory(dirName):
    return db().DeleteDirectory(dirName)
    
def CheckDirectory(dirName):
    return db().CheckDirectory(dirName)
    
def GetBackupDirectories():
    return db().GetBackupDirectories()
    
def GetDirectoryInfo(dirName):
    return db().GetDirectoryInfo(dirName)

def AddDirBackup(dirName, backupID, backupStatus="", backupSize=0, backupStart=0, backupFinish=0):
    return db().AddDirBackup(dirName, backupID, backupStatus, backupSize, backupStart, backupFinish)
    
def DeleteDirBackup(backupID):
    return db().DeleteDirBackup(backupID)
    
def IsDeleted(backupID):
    return db().IsDeleted(backupID)
    
def SetBackupStatus(dirName, backupID, backupStatus, backupFinish):
    return db().SetBackupStatus(dirName, backupID, backupStatus, backupFinish)
    
def AbortDirectoryBackup(dirName):
    return db().AbortDirectoryBackup(dirName)
    
def GetLastRunInfo(dirName):
    return db().GetLastRunInfo(dirName)
    
def IsBackupRunning(dirName):
    return db().IsBackupRunning(dirName)

def GetSchedule(dirName):
    return db().GetDirSchedule(dirName)

def SetSchedule(dirName, sched):
    return db().SetDirSchedule(dirName, sched)

def GetLastRun(dirName):
    return db().GetDirLastBackup(dirName)

def GetDirBackupIds(dirName):
    return db().GetDirBackupIds(dirName)

def GetBackupIds(full_info_please=False):
    return db().GetBackupIds(full_info_please)
    
def GetTotalBackupsSize():
    return db().GetTotalBackupsSize()

def GetDirectoryFromBackupId(backupID):
    return db().GetDirectoryFromBackupId(backupID)

def GetDirectorySubfoldersInclude(dirName):
    return db().GetDirectorySubfoldersInclude(dirName)

def SetDirectorySubfoldersInclude(dirName, backupSubfolders):
    return db().SetDirectorySubfoldersInclude(dirName, backupSubfolders)

def GetBackupIdRunInfo(backupID):
    return db().GetBackupIdRunInfo(backupID)

def GetBackupsByDateTime(reverse=False):
    return db().GetBackupsByDateTime(reverse)
    
def GetBackupsByFolder():
    return db().GetBackupsByFolder()

def AddRunningBackupObject(backupID, backupObject):
    return db().AddRunningBackupObject(backupID, backupObject)

def GetRunningBackupObject(backupID):
    return db().GetRunningBackupObject(backupID)

def RemoveRunningBackupObject(backupID):
    return db().RemoveRunningBackupObject(backupID)

def RemoveAllRunningBackupObjects():
    return db().RemoveAllRunningBackupObjects()

def HasRunningBackup():
    return db().HasRunningBackup()

def ShowRunningBackups():
    return db().ShowRunningBackups()

def AbortRunningBackup(backupID):
    return db().AbortRunningBackup(backupID)

#------------------------------------------------------------------------------ 

def db():
    global _DB
    if _DB is None:
        _DB = _BackupDB()
    return _DB

def init():
    _local_backup_db = db()

    global InitDone
    if InitDone:
        return

    dhnio.Dprint(4,"backup_db.init")

    _local_backup_db.Load()
    _local_backup_db.AskSizeForAllDirs()

    #init state
    InitDone = True


#------------------------------------------------------------------------------ 

if __name__ == "__main__":
    init()
    import pprint
    pprint.pprint(GetBackupIds())
    pprint.pprint(GetBackupDirectories())
#    newdir = unicode(sys.argv[1])
#    print 'want to add:', newdir
##    AddDirBackup(newdir, "F3000000000AM", "done", "12344321", "1234554321.11", "1234554323.11")
#    AddDirectory(newdir)
##    DeleteDirectory(newdir)
#    print ToXML()
##    print GetDirectoryFromBackupId("F3000000000AM")
##    print CheckDirectory(newdir)
    #_local_backup_db.Save()



#!/usr/bin/python
#dhnio.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2013
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#
# This is for simple routines that do not require importing any of our code

import os
import sys
import time
import imp
import string
import platform
import traceback
import locale
import threading
import glob
import re

#------------------------------------------------------------------------------

# Code will set 1 for most important and 10 for really minor stuff, and 15 for things we don't think we want to see again
# Can set DebugLevel to 0 for no debug messages
DebugLevel = 0
LogLinesCounter = 0
EnableLog = True
RedirectStdOut = False
NoOutput = False
OriginalStdOut = None
StdOutPrev = None
LogFile = None
LogFileName = None
WebStreamFunc = None
ShowTime = True
LifeBeginsTime = 0
tc_using = True
tc_remove_after = True
LocaleInstalled = False
PlatformInfo = None
X11isRunning = None

#------------------------------------------------------------------------------

def init():
    InstallLocale()
    if Linux():
        UnbufferedSTDOUT()
    # StartCountingOpenedFiles()
        
def shutdown():
    Dprint(2, 'dhnio.shutdown')
    RestoreSTDOUT()

def InstallLocale():
    global LocaleInstalled
    if LocaleInstalled:
        return False
    try:
        import sys
        reload(sys)
        if hasattr(sys, "setdefaultencoding"):
            import locale
            denc = locale.getpreferredencoding()
            if denc != '':
                sys.setdefaultencoding(denc)
        LocaleInstalled = True
    except:
        pass
    return LocaleInstalled

# from here: 
# http://algorithmicallyrandom.blogspot.com/2009/10/python-tips-and-tricks-flushing-stdout.html
# Thanks!!!
def UnbufferedSTDOUT():
    global OriginalStdOut
    OriginalStdOut = sys.stdout
    sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)

def RestoreSTDOUT():
    global OriginalStdOut
    try:
        dhn_std_out = sys.stdout
        sys.stdout = OriginalStdOut
        dhn_std_out.close()
    except:
        traceback.print_last(file=open('dhnio.shutdown.error', 'w'))

# http://habrahabr.ru/post/117236/
# import sys
# import codecs
def setup_console(sys_enc="utf-8"):
    reload(sys)
    try:
        if sys.platform.startswith("win"):
            import ctypes
            enc = "cp%d" % ctypes.windll.kernel32.GetOEMCP()
        else:
            enc = (sys.stdout.encoding if sys.stdout.isatty() else
                        sys.stderr.encoding if sys.stderr.isatty() else
                            sys.getfilesystemencoding() or sys_enc)

        sys.setdefaultencoding(sys_enc)

        if sys.stdout.isatty() and sys.stdout.encoding != enc:
            sys.stdout = codecs.getwriter(enc)(sys.stdout, 'replace')

        if sys.stderr.isatty() and sys.stderr.encoding != enc:
            sys.stderr = codecs.getwriter(enc)(sys.stderr, 'replace')

    except:
        pass


#  Also, sys.platform is things like "win32" and "linux2"
#  also  os.name is "posix" for linux, "nt" for "nt" and ____ for windows
#  "Linux", "Windows"
#  don't print anything in ostype because used in dhnbackup.py and stdout goes to tar file
def ostype():
    global PlatformInfo
    if PlatformInfo is None:
        PlatformInfo = platform.uname()
    return PlatformInfo[0]

def osversion():
    global PlatformInfo
    if PlatformInfo is None:
        PlatformInfo = platform.uname()
    return PlatformInfo[2]

def osinfo():
    return str(platform.platform()).strip()

def osinfofull():
    import pprint
    o = ''
    o += '=====================================================\n'
    o += '=====================================================\n'
    o += '=====================================================\n'
    o += 'platform.uname(): ' + str(platform.uname()) + '\n'
    o += 'sys.executable: ' + sys.executable + '\n'
    o += 'os.path.abspath("."): ' + os.path.abspath('.') + '\n'
    o += 'os.path.abspath(sys.argv[0]): ' + os.path.abspath(sys.argv[0]) + '\n'
    o += 'os.path.expanduser("~"): ' + os.path.expanduser('~') + '\n'
    o += 'sys.argv: ' + pprint.pformat(sys.argv) + '\n'
    o += 'sys.path:\n' + pprint.pformat(sys.path) + '\n'
    o += 'os.environ:\n' + pprint.pformat(os.environ.items()) + '\n'
    o += '=====================================================\n'
    o += '=====================================================\n'
    o += '=====================================================\n'
    return o

def windows_version():
    if getattr(sys, 'getwindowsversion', None) is not None:
        return sys.getwindowsversion()[0]
    return 0

def Linux():
    return ostype() == "Linux"

def Windows():
    return ostype() == "Windows"

def Mac():
    return ostype() == "Darwin"

def isFrozen():
    return main_is_frozen()

def FrozenIsDebug():
    if not main_is_frozen():
        return False
    return os.path.basename(sys.executable) == 'dhnmaind.exe'

def SetDebug(level):
    global DebugLevel
    if DebugLevel > level:
        Dprint(level, "dhnio.SetDebug DebugLevel=" + str(level))
    DebugLevel = level

def LifeBegins():
    global LifeBeginsTime
    LifeBeginsTime = time.time()

# return True if something at this level should be reported given current DebugLevel
def Debug(level):
    global DebugLevel
    return level <= DebugLevel

def Dprint(level, s, nl='\n'):
    global WebStreamFunc
    global LogFile
    global RedirectStdOut
    global ShowTime
    global LifeBeginsTime
    global NoOutput
    global LogLinesCounter
    global EnableLog
    global DebugLevel
    if not EnableLog:
        return
    s_ = s
    if level % 2:
        level -= 1
    if level:
        s = ' ' * level + s
    if ShowTime and level > 0:
        if LifeBeginsTime != 0:
            dt = time.time() - LifeBeginsTime
            mn = dt // 60
            sc = dt - mn * 60
            if DebugLevel>=8:
                s = ('%02d:%02d.%02d' % (mn, sc, (sc-int(sc))*100)) + s
            else:
                s = ('%02d:%02d' % (mn, sc)) + s
        else:
            s = time.strftime('%H:%M:%S') + s
    if Debug(18):
        currentThreadName = threading.currentThread().getName()
        # if currentThreadName.find('MainThread') >= 0:
        s = s + ' {%s}' % currentThreadName.lower()
    if Debug(level):
        if LogFile is not None:
            LogFile.write(s + nl)
            LogFile.flush()
        #else:
        if not RedirectStdOut and not NoOutput:
            # sys.stdout.write(s + nl)
            #print s
            if nl == '\n':
                print s
            else:
                sys.stdout.write(s + nl)
    if WebStreamFunc is not None:
        #WebStreamFunc(level, s_.rstrip())
        WebStreamFunc(level, s_ + nl)
    LogLinesCounter += 1
    if LogLinesCounter % 10000 == 0:
        Dprint(2, '[%s]' % time.asctime())

def DprintStd(level, s):
    s_ = s
    level_ = level
    if level_ % 2:
        level_ -= 1
    if level == 1:
        level_ = 1
    if level_:
        s_ = ' ' * level_ + s_
        s_ = time.strftime('%H:%M:%S') + s_
    print s_

def Dglobals(level, glob_dict):
    global DebugLevel
    if (level>DebugLevel):
        return
    keys = glob_dict.keys()
    keys.sort()
    for k in keys:
        if k != '__builtins__':
            print k, glob_dict[k]

_TimePush = 0.0
_TimeTotalDict = {}
_TimeDeltaDict = {}
_TimeCountsDict = {}
def DTimePush(t):
    global _TimeTotalDict
    global _TimeDeltaDict
    global _TimeCountsDict
    tm = time.time()
    if not _TimeTotalDict.has_key(t):
        _TimeTotalDict[t] = 0.0
        _TimeCountsDict[t] = 0
    _TimeDeltaDict[t] = tm

def DTimePop(t):
    global _TimeTotalDict
    global _TimeDeltaDict
    global _TimeCountsDict
    tm = time.time()
    if not _TimeTotalDict.has_key(t):
        return
    dt = tm - _TimeDeltaDict[t]
    _TimeTotalDict[t] += dt
    _TimeCountsDict[t] += 1

def DprintTotalTime():
    global _TimeTotalDict
    global _TimeDeltaDict
    global _TimeCountsDict
    for t in _TimeTotalDict.keys():
        total = _TimeTotalDict[t]
        counts = _TimeCountsDict[t]
        Dprint(2, 'total=%f sec. count=%d, avarage=%f: %s' % (total, counts, total/counts, t))

def exceptionName(value):
    try:
        excStr = unicode(value)
    except:
        try:
            excStr = repr(value)
        except:
            try:
                excStr = str(value)
            except:
                try:
                    excStr = value.message
                except:
                    excStr = type(value).__name__
    return excStr

def formatExceptionInfo(maxTBlevel=100, exc_info=None):
    if exc_info is None:
        cla, value, trbk = sys.exc_info()
    else:
        cla, value, trbk = exc_info
    try:
        excArgs = value.__dict__["args"]
    except KeyError:
        excArgs = ''
    excTb = traceback.format_tb(trbk, maxTBlevel)
    tbstring = 'Exception: <' + exceptionName(value) + '>\n'
    if excArgs:
        s += '  args:' + excArgs + '\n' 
    for s in excTb:
        tbstring += s + '\n'
    return tbstring

def DprintException(level=0, maxTBlevel=100, exc_info=None):
    global LogFileName
    if exc_info is None:
        cla, value, trbk = sys.exc_info()
    else:
        cla, value, trbk = exc_info
    try:
        excArgs = str(value.__dict__["args"])
    except KeyError:
        excArgs = ''
    excTb = traceback.format_tb(trbk, maxTBlevel)
    s = 'Exception: <' + exceptionName(value) + '>\n'
    Dprint(level, s.strip())
    if excArgs:
        s += '  args:' + excArgs + '\n'
        Dprint(level, '  args:' + excArgs)
    s += '\n'
    excTb.reverse()
    for l in excTb:
        Dprint(level, l.replace('\n', ''))
        s += l + '\n'
    try:
        file = open(os.path.join(os.path.dirname(LogFileName), 'exception.log'), 'w')
        file.write(s)
        file.close()
    except:
        pass

def ExceptionHook(type, value, traceback):
    DprintException(exc_info=(type, value, traceback))

#------------------------------------------------------------------------------

def SetWebStream(webstreamfunc):
    global WebStreamFunc
    WebStreamFunc = webstreamfunc

def WebStream():
    global WebStreamFunc
    return WebStreamFunc

def OpenLogFile(filename, append_mode=False):
    global LogFile
    global LogFileName
    if LogFile:
        return
    try:
        if not os.path.isdir(os.path.dirname(os.path.abspath(filename))):
            os.makedirs(os.path.dirname(os.path.abspath(filename)))
        if append_mode:
            LogFile = open(os.path.abspath(filename), 'a')
        else:
            LogFile = open(os.path.abspath(filename), 'w')
        #sys.stderr = LogFile
        #sys.stdout = LogFile
        LogFileName = os.path.abspath(filename)
    except:
        Dprint(0, 'cant open ' + filename)
        DprintException()

def CloseLogFile():
    global LogFile
    if not LogFile:
        return
    LogFile.flush()
    LogFile.close()
    LogFile = None

class DHN_stdout:
    softspace = 0
    def read(self): pass
    def write(self, s):
        # Dprint(0, s.rstrip())
        Dprint(0, unicode(s).rstrip())
    def flush(self): pass
    def close(self): pass

class DHN_black_hole:
    softspace = 0
    def read(self): pass
    def write(self, s):  pass
    def flush(self): pass
    def close(self): pass


def StdOutRedirectingStart():
    global RedirectStdOut
    global StdOutPrev
    RedirectStdOut = True
    StdOutPrev = sys.stdout
    sys.stdout = DHN_stdout()

def StdOutRedirectingStop():
    global RedirectStdOut
    global StdOutPrev
    RedirectStdOut = False
    if StdOutPrev is not None:
        sys.stdout = StdOutPrev

def DisableOutput():
    global RedirectStdOut
    global StdOutPrev
    global NoOutput
    NoOutput = True
    RedirectStdOut = True
    StdOutPrev = sys.stdout
    sys.stdout = DHN_black_hole()

def DisableLogs():
    global EnableLog
    EnableLog = False

#-------------------------------------------------------------------------------

def list_dir_recursive(dir):
    r = []
    for name in os.listdir(dir):
        full_name = os.path.join(dir, name)
        if os.path.isdir(full_name):
            r.extend(list_dir_recursive(full_name))
        else:
            r.append(full_name)
    return r

#http://mail.python.org/pipermail/python-list/2000-December/060960.html
def rmdir_recursive(dirpath, ignore_errors=False):
    """Remove a directory, and all its contents if it is not already empty."""
    for name in os.listdir(dirpath):
        full_name = os.path.join(dirpath, name)
        # on Windows, if we don't have write permission we can't remove
        # the file/directory either, so turn that on
        if not os.access(full_name, os.W_OK):
            os.chmod(full_name, 0600)
        if os.path.isdir(full_name):
            rmdir_recursive(full_name)
        else:
            if not ignore_errors:
                os.remove(full_name)
            else:
                try:
                    os.remove(full_name)
                except:
                    Dprint(6, 'dhnio.rmdir_recursive can not remove file ' + full_name)

    if not ignore_errors:
        os.rmdir(dirpath)
    else:
        try:
            os.rmdir(dirpath)
        except:
            Dprint(6, 'dhnio.rmdir_recursive can not remove dir ' + dirpath)

def getDirectorySize(directory, include_subfolders=True):
    if Windows():
        import win32file
        import win32con
        import pywintypes
        DIR_EXCLUDES = set(['.', '..'])
        MASK = win32con.FILE_ATTRIBUTE_DIRECTORY | win32con.FILE_ATTRIBUTE_SYSTEM
        REQUIRED = win32con.FILE_ATTRIBUTE_DIRECTORY
        FindFilesW = win32file.FindFilesW
        def _get_dir_size(path):
            total_size = 0
            try:
                items = FindFilesW(path + r'\*')
            except pywintypes.error, ex:
                return total_size
            for item in items:
                total_size += item[5]
                if item[0] & MASK == REQUIRED and include_subfolders:
                    name = item[8]
                    if name not in DIR_EXCLUDES:
                        total_size += _get_dir_size(path + '\\' + name)
            return total_size
        return _get_dir_size(directory)

    dir_size = 0
    if  not include_subfolders:
        for filename in os.listdir(directory):
            filepath = os.path.abspath(os.path.join(directory, filename))
            if os.path.isfile(filepath):
                try:
                    dir_size += os.path.getsize(filepath)
                except:
                    pass
    else:
        for (path, dirs, files) in os.walk(directory):
            for file in files:
                filename = os.path.join(path, file)
                if os.path.isfile(filename):
                    try:
                        dir_size += os.path.getsize(filename)
                    except:
                        pass
    return dir_size

def getHomeDirectory():
    if not Windows():
        return os.path.expanduser('~')
    try:
        from win32com.shell import shell
        df = shell.SHGetDesktopFolder()
        pidl = df.ParseDisplayName(0, None,
            "::{450d8fba-ad25-11d0-98a8-0800361b1103}")[1]
        mydocs = shell.SHGetPathFromIDList(pidl)
##        another one method:
##        import win32com.client
##        objShell = win32com.client.Dispatch("WScript.Shell")
##        mydocs = objShell.SpecialFolders("MyDocuments")
##             Other available folders:
##             AllUsersDesktop, AllUsersStartMenu, AllUsersPrograms,
##             AllUsersStartup, Desktop, Favorites, Fonts, MyDocuments,
##             NetHood, PrintHood, Recent, SendTo, StartMenu, Startup & Templates
        return mydocs
    except:
        return os.path.expanduser('~')


#-------------------------------------------------------------------------------


### AtomicSave:  Save either all of data to file, or don't make file
##def AtomicSave(filename, data):
##    tmp = '%s.tmp' % filename
##    f = file(tmp, 'wb')
##    f.write(data)
##    os.fsync(f)
##    f.close()
##    os.rename(tmp, filename)

def AtomicWriteFile(filename, data):
    try:
        tmpfilename = filename + ".new"
        f = open(tmpfilename, "wb")
        f.write(data)
        f.flush()
        #from http://docs.python.org/library/os.html on os.fsync
        os.fsync(f.fileno())
        f.close()
        #in Unix the rename will overwrite an existing file,
        #but in Windows it fails, so have to remove existing
        if Windows() and os.path.exists(filename):
            os.remove(filename)
        os.rename(tmpfilename, filename)
    except:
        Dprint(1, 'dhnio.AtomicWriteFile ERROR ' + str(filename))
        DprintException()
        try:
            f.close() # make sure file gets closed
        except:
            pass
        return False
    return True

def AtomicAppendFile(filename, data, mode='a'): # TODO, this is not atomic
    try:
        f = open(filename, mode)
        f.write(data)
        f.flush()
        os.fsync(f.fileno())
        f.close()
    except:
        Dprint(1, 'dhnio.AtomicAppendFile ERROR ' + str(filename))
        DprintException()
        try:
            f.close() # make sure file gets closed
        except:
            DprintException()
        return False
    return True

# PREPRO - probably all writes should be Atomic, so we should write to temp file then rename
def WriteFile(filename, data):
    return AtomicWriteFile(filename, data)

def WriteFileSimple(filename, data, mode="w"):
    try:
        file = open(filename, mode)
        file.write(data)
        file.close()
    except:
        DprintException()
        return False
    return True

def ReadTextFile(filename):
    if not os.path.isfile(filename):
        return ''
    if not os.access(filename, os.R_OK):
        return ''
    try:
        file=open(filename,"r")
        data=file.read()
        file.close()
        # Windows/Linux trouble with text files
        return data.replace('\r\n','\n')
    except:
        DprintException()
        return ''

def ReadBinaryFile(filename):
    if not os.path.isfile(filename):
        return ''
    if not os.access(filename, os.R_OK):
        return ''
    try:
        file = open(filename,"rb")
        data = file.read()
        file.close()
        return data
    except:
        DprintException()
        return ''

#-------------------------------------------------------------------------------

def _read_data(path):
    if not os.path.exists(path):
        return None
    if not os.access(path, os.R_OK):
        return None
    fin = open(path, 'r')
    src = fin.read()
    fin.close()
    return src

def _write_data(path, src):
    temp_path = path + '.tmp'
    if os.path.exists(temp_path):
        if not os.access(temp_path, os.W_OK):
            return False
    if os.path.exists(path):
        if not os.access(path, os.W_OK):
            return False
        try:
            os.remove(path)
        except:
            Dprint(1, 'dhnio._write_data ERROR removing ' + str(path))
    fout = open(temp_path, 'wb')
    fout.write(src)
    fout.flush()
    os.fsync(fout)
    fout.close()
    try:
        os.rename(temp_path, path)
    except:
        Dprint(1, 'dhnio._write_data ERROR renaming %s to %s' % (str(temp_path), str(path)))
    return True

def _append_data(path, src):
    if os.path.exists(path):
        if not os.access(path, os.W_OK):
            return False
    fout = open(path, 'a')
    fout.write(src)
    fout.flush()
    os.fsync(fout)
    fout.close()
    return True

#-------------------------------------------------------------------------------

def _pack_list(lst):
    return str(len(lst))+'\n'+'\n'.join(lst)

def _unpack_list(src):
    if src.strip() == '':
        return list(), None
    words = src.splitlines()
    if len(words) == 0:
        return list(), None
    try:
        length = int(words[0])
    except:
        return words, None
    res = words[1:]
    if len(res) < length:
        res += [''] * (length - len(res))
    elif len(res) > length:
        return res[:length], res[length:]
    return res, None

def _read_list(path):
    src = _read_data(path)
    if src is None:
        return None
    return _unpack_list(src)[0]

def _write_list(path, lst):
    return _write_data(path, _pack_list(lst))

#-------------------------------------------------------------------------------

def _pack_dict(dictionary, sort=False):
    src = ''
    if sort:
        for k in sorted(dictionary.keys()):
            src += k + ' ' + str(dictionary[k]) + '\n'
    else:
        for k in sorted(dictionary.keys()):
            src += k + ' ' + str(dictionary[k]) + '\n'
    return src

def _unpack_dict_from_list(lines):
    dct = {}
    for line in lines:
        words = line.split(' ')
        if len(words) < 2:
            continue
        dct[words[0]] = ' '.join(words[1:])
    return dct

def _unpack_dict(src):
    lines = src.split('\n')
    return _unpack_dict_from_list(lines)

def _read_dict(path, default=None):
    src = _read_data(path)
    if src is None:
        return default
    return _unpack_dict(src.strip())

def _write_dict(path, dictionary, sort=False):
    data = _pack_dict(dictionary, sort)
    return _write_data(path, data)

def _dir_exist(path):
    return os.path.isdir(path)

def _dir_make(path):
    os.mkdir(path, 0777)

def _dirs_make(path):
    os.makedirs(path, 0777)

def _dir_remove(path):
    rmdir_recursive(path)

def backup_and_remove(path):
    bkpath = path + '.backup'
    if not os.path.exists(path):
        return
    if os.path.exists(bkpath):
        try:
            os.remove(bkpath)
        except:
            Dprint(1, 'dhnio.backup_and_remove ERROR can not remove file ' + bkpath)
            DprintException()

    try:
        os.rename(path, bkpath)
    except:
        Dprint(1, 'dhnio.backup_and_remove ERROR can not rename file %s to %s' % (path, bkpath))
        DprintException()

    if os.path.exists(path):
        try:
            os.remove(path)
        except:
            Dprint(1, 'dhnio.backup_and_remove ERROR can not remove file ' + path)
            DprintException()

def restore_and_remove(path, overwrite_existing = False):
    bkpath = path + '.backup'
    if not os.path.exists(bkpath):
        return

    if os.path.exists(path):
        if not overwrite_existing:
            return
        try:
            os.remove(path)
        except:
            Dprint(1, 'dhnio.restore_and_remove ERROR can not remove file ' + path)
            DprintException()

    try:
        os.rename(bkpath, path)
    except:
        Dprint(1, 'dhnio.restore_and_remove ERROR can not rename file %s to %s' % (path, bkpath))
        DprintException()

def remove_backuped_file(path):
    bkpath = path + '.backup'
    if not os.path.exists(bkpath):
        return
    try:
        os.remove(bkpath)
    except:
        Dprint(1, 'dhnio.remove_backuped_file ERROR can not remove file ' + bkpath)
        DprintException()

#  Lower the priority of the running process
def LowerPriority():
    try:
        sys.getwindowsversion()
    except:
        isWindows = False
    else:
        isWindows = True

    if isWindows:
        # Based on:
        #   "Recipe 496767: Set Process Priority In Windows" on ActiveState
        #   http://code.activestate.com/recipes/496767/
        import win32api
        import win32process
        import win32con
        pid = win32api.GetCurrentProcessId()
        handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
        win32process.SetPriorityClass(handle, win32process.BELOW_NORMAL_PRIORITY_CLASS)
    else:
        import os
        os.nice(20)

#-------------------------------------------------------------------------------

def shortPath(path):
    #get Absolute Short Path in Unicode
    #convert to 8.3 windows filenames
    path_ = os.path.abspath(path)
    if not Windows():
        if type(path_) != unicode:
            return unicode(path_)
        return path_

    if not os.path.exists(path_):
        if os.path.isdir(os.path.dirname(path_)):
            res = shortPath(os.path.dirname(path_))
            return unicode(os.path.join(res, os.path.basename(path_)))
        return unicode(path_)

    try:
        import win32api
        spath = win32api.GetShortPathName(path_)
        return unicode(spath)
    except:
        DprintException()
        return unicode(path_)

def longPath(path):
    #get Absolute Long Path in Unicode
    #convert to full path. even if it was in 8.3 format
    path_ = os.path.abspath(path)
    if not Windows():
        if type(path_) != unicode:
            return unicode(path_)
        return path_

    if not os.path.exists(path_):
        return unicode(path_)

    try:
        import win32api
        lpath = win32api.GetLongPathName(path_)
        return unicode(lpath)
    except:
        DprintException()
    return unicode(path_)

#------------------------------------------------------------------------------ 

# http://www.py2exe.org/index.cgi/HowToDetermineIfRunningFromExe
def main_is_frozen():
    return (hasattr(sys, "frozen") or # new py2exe
            hasattr(sys, "importers") or# old py2exe
            imp.is_frozen("__main__")) # tools/freeze

def X11_is_running():
    global X11isRunning
    if not Linux():
        return False
    if X11isRunning is not None:
        return X11isRunning
    try:
        # http://stackoverflow.com/questions/1027894/detect-if-x11-is-available-python
        from subprocess import Popen, PIPE
        p = Popen(["xset", "-q"], stdout=PIPE, stderr=PIPE)
        p.communicate()
        result = p.returncode == 0
    except:
        result = False
    X11isRunning = result 
    return X11isRunning

#------------------------------------------------------------------------------ 

def getExecutableDir():
    if main_is_frozen():
        path = os.path.dirname(os.path.abspath(sys.executable))
    else:
        path = os.path.dirname(os.path.abspath(sys.argv[0]))
#    if Windows():
#        return shortPath(path)
    return unicode(path)

def getExecutableFilename():
    if main_is_frozen():
        path = os.path.abspath(sys.executable)
    else:
        path = os.path.abspath(sys.argv[0])
#    if Windows():
#        return shortPath(path)
    return unicode(path)

def getExecutableP2PDir():
    execdir = getExecutableDir()
    if os.path.isdir(os.path.join(execdir, 'p2p')):
        return os.path.join(execdir, 'p2p')
    return execdir

def getUserName():
    return os.path.basename(unicode(os.path.expanduser('~')))

#------------------------------------------------------------------------------

def listHomeDirLinux():
    if Windows():
        return []
    rootlist = []
    homedir = os.path.expanduser('~')
    for dirname in os.listdir(homedir):
        if os.path.isdir(os.path.join(homedir, dirname)):
            rootlist.append(dirname)
    return rootlist

def listLocalDrivesWindows():
    if not Windows():
        return []
    rootlist = []
    try:
        import win32api
        import win32file
        drives = (drive for drive in win32api.GetLogicalDriveStrings().split ("\000") if drive)
        for drive in drives:
            if win32file.GetDriveType(drive) == 3:
                rootlist.append(drive)
    except:
        DprintException()
    return rootlist

# def listRemovableDrivesWindowsOld():
#     l = []
#    try:
#        import wmi
#        c = wmi.WMI()
#        for d in c.Win32_LogicalDisk(DriveType=2):
#            l.append(d.Name+'\\')
#        del c
#    except:
#        DprintException()
#     return l

def listRemovableDrivesWindows():
    l = []
    try:
        import win32file
        drivebits = win32file.GetLogicalDrives()
        for d in range(1, 26):
            mask = 1 << d
            if drivebits & mask:
                # here if the drive is at least there
                drname='%c:\\' % chr(ord('A')+d)
                t = win32file.GetDriveType(drname)
                # print drname, t
                # drive_type = ctypes.windll.kernel32.GetDriveTypeA(drname)                    
                if t == win32file.DRIVE_REMOVABLE:
                    l.append(drname)
    except:
        DprintException()
    return l

def listRemovableDrivesLinux():
    try:
        return map(lambda x: os.path.join('/media', x), os.listdir('/media'))
    except:
        return []

def listRemovableDrives():
    if Linux():
        return listRemovableDrivesLinux()
    elif Windows():
        return listRemovableDrivesWindows()
    return []

def listMountPointsLinux():
    mounts = os.popen('mount')
    result = []
    for line in mounts.readlines():
        mo = re.match('^(.+?) on (.+?) type .+?$', line)
        if mo:
            device = mo.group(1)
            mount_point = mo.group(2)
            if device.startswith('/dev/'):
                result.append(mount_point)
    return result
        
def getMountPointLinux(path):
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path

#-------------------------------------------------------------------------------

def find_process(applist):
    pidsL = []
    ostype = platform.uname()[0]
    if ostype == "Windows":
        return find_process_win32(applist)
    else:
        return find_process_linux(applist)
    return pidsL


def kill_process(pid):
    ostype = platform.uname()[0]
    if ostype == "Windows":
        kill_process_win32(pid)
    else:
        kill_process_linux(pid)


def list_processes_linux():
    """
    This function will return an iterator with the process pid/cmdline tuple

    :return: pid, cmdline tuple via iterator
    :rtype: iterator

    >>> for procs in list_processes():
    >>>     print procs
    ('5593', '/usr/lib/mozilla/kmozillahelper')
    ('6353', 'pickup -l -t fifo -u')
    ('6640', 'kdeinit4: konsole [kdeinit]')
    ('6643', '/bin/bash')
    ('7451', '/usr/bin/python /usr/bin/ipython')
    """
    for pid_path in glob.glob('/proc/[0-9]*'):
        try:
            # cmdline represents the command whith which the process was started
            f = open("%s/cmdline" % pid_path)
            pid = pid_path.split("/")[2] # get the PID
            # we replace the \x00 to spaces to make a prettier output from kernel
            cmdline = f.read().replace("\x00", " ").rstrip()
            f.close()
    
            yield (pid, cmdline)
        except:
            pass


def find_process_linux(applist):
    pidsL = []
    for pid, cmdline in list_processes_linux():
        try:
            pid = int(pid)
        except:
            continue
        if pid == os.getpid():
            continue
        for app in applist:
#            if isinstance(app, list):
#                found = True
#                for word in app:
#                    if cmdline.find(word) == -1:
#                        found = False
#                        break
#                if found:
#                    pidsL.append(pid)
#            else:
            if app.startswith('regexp:'):
                if re.match(app[7:], cmdline) is not None:
                    pidsL.append(pid)
            else:
                if cmdline.find(app) > -1:
                    pidsL.append(pid)
    return pidsL


def find_process_win32(applist):
    pidsL = []
    try:
        import win32com.client
        objWMI = win32com.client.GetObject("winmgmts:\\\\.\\root\\CIMV2")
        colProcs = objWMI.ExecQuery("SELECT * FROM Win32_Process")
        for Item in colProcs:
            pid = int(Item.ProcessId)
            if pid == os.getpid():
                continue
            cmdline = Item.Caption.lower()
            if Item.CommandLine:
                cmdline += Item.CommandLine.lower()
            for app in applist:
#                if isinstance(app, list):
#                    found = True
#                    for word in app:
#                        if cmdline.find(word) == -1:
#                            found = False
#                            break
#                    if found:
#                        pidsL.append(pid)
#                else:
                if app.startswith('regexp:'):
                    if re.match(app[7:], cmdline) is not None:
                        pidsL.append(pid)
                else:
                    if cmdline.find(app) > -1:
                        pidsL.append(pid)
                    
    except:
        DprintException()
    return pidsL


def kill_process_linux(pid):
    try:
        import signal
        os.kill(pid, signal.SIGTERM)
    except:
        DprintException()


def kill_process_win32(pid):
    try:
        from win32api import TerminateProcess, OpenProcess, CloseHandle
    except:
        DprintException()
        return False
    try:
        PROCESS_TERMINATE = 1
        handle = OpenProcess(PROCESS_TERMINATE, False, pid)
    except:
        Dprint(2, 'dhnio.kill_process_win32 can not open process %d' % pid)
        return False
    try:
        TerminateProcess(handle, -1)
    except:
        Dprint(2, 'dhnio.kill_process_win32 can not terminate process %d' % pid)
        return False
    try:
        CloseHandle(handle)
    except:
        DprintException()
        return False
    return True



    




#!/usr/bin/env python
#initializer.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2013
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#

import os
import sys

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in initializer.py')
from twisted.internet.defer import Deferred, maybeDeferred
from twisted.internet.task import LoopingCall

import lib.dhnio as dhnio
# from lib.automat import Automat
import lib.automat as automat
import lib.automats as automats

import network_connector
import installer
import shutdowner
import p2p_connector

import dhninit
import webcontrol

_Initializer = None

#------------------------------------------------------------------------------

def A(event=None, arg=None, use_reactor=True):
    global _Initializer
    if _Initializer is None:
        _Initializer = Initializer('initializer', 'AT_STARTUP', 2)
    if event is not None:
        if use_reactor:
            _Initializer.automat(event, arg)
        else:
            _Initializer.event(event, arg)
    return _Initializer


class Initializer(automat.Automat):
    
    def init(self):
        self.flagCmdLine = False
        self.is_installed = None
    
    def state_changed(self, oldstate, newstate):
        automats.set_global_state('INIT ' + newstate)

    def A(self, event, arg):
        #---AT_STARTUP---
        if self.state is 'AT_STARTUP':
            if event == 'run' :
                self.state = 'LOCAL'
                self.doInitLocal(arg)
                self.flagCmdLine=False
                shutdowner.A('init')
            elif event == 'run-cmd-line-register' :
                self.state = 'INSTALL'
                self.flagCmdLine=True
                installer.A('register-cmd-line', arg)
                shutdowner.A('init')
                shutdowner.A('ready')
            elif event == 'run-cmd-line-recover' :
                self.state = 'INSTALL'
                self.flagCmdLine=True
                installer.A('recover-cmd-line', arg)
                shutdowner.A('init')
                shutdowner.A('ready')
        #---LOCAL---
        elif self.state is 'LOCAL':
            if event == 'init-local-done' and self.isInstalled(arg) :
                self.state = 'CONTACTS'
                self.doInitContacts(arg)
            elif event == 'init-local-done' and not self.isInstalled(arg) and self.isGUIPossible(arg) :
                self.state = 'INSTALL'
                self.doShowGUI(arg)
                installer.A('init')
                self.doUpdate(arg)
                shutdowner.A('ready')
            elif event == 'init-local-done' and not self.isInstalled(arg) and not self.isGUIPossible(arg) :
                self.state = 'STOPPING'
                shutdowner.A('stop')
        #---CONTACTS---
        elif self.state is 'CONTACTS':
            if event == 'init-contacts-done' :
                self.state = 'CONNECTION'
                self.doInitConnection(arg)
                network_connector.A('init')
                p2p_connector.A('init') ;
                self.doShowGUI(arg)
                self.doUpdate(arg)
                shutdowner.A('ready')
            elif ( event == 'shutdowner.state' and arg is 'FINISHED' ) :
                self.state = 'EXIT'
                self.doDestroyMe(arg)
        #---CONNECTION---
        elif self.state is 'CONNECTION':
            if ( event == 'shutdowner.state' and arg is 'FINISHED' ) :
                self.state = 'EXIT'
                self.doDestroyMe(arg)
            elif ( event == 'p2p_connector.state' and arg in [ 'CONNECTED' , 'DISCONNECTED' ] ) :
                self.state = 'MODULES'
                self.doInitModules(arg)
                self.doUpdate(arg)
        #---MODULES---
        elif self.state is 'MODULES':
            if event == 'init-modules-done' :
                self.state = 'READY'
                self.doUpdate(arg)
            elif ( event == 'shutdowner.state' and arg is 'FINISHED' ) :
                self.state = 'EXIT'
                self.doDestroyMe(arg)
        #---INSTALL---
        elif self.state is 'INSTALL':
            if not self.flagCmdLine and ( event == 'installer.state' and arg is 'DONE' ) :
                self.state = 'CONTACTS'
                self.doInitContacts(arg)
                self.doUpdate(arg)
            elif self.flagCmdLine and ( event == 'installer.state' and arg is 'DONE' ) :
                self.state = 'STOPPING'
                shutdowner.A('stop')
            elif ( event == 'shutdowner.state' and arg is 'FINISHED' ) :
                self.state = 'EXIT'
                self.doDestroyMe(arg)
        #---READY---
        elif self.state is 'READY':
            if ( event == 'shutdowner.state' and arg is 'FINISHED' ) :
                self.state = 'EXIT'
                self.doDestroyMe(arg)
        #---STOPPING---
        elif self.state is 'STOPPING':
            if ( event == 'shutdowner.state' and arg is 'FINISHED' ) :
                self.state = 'EXIT'
                self.doUpdate(arg)
                self.doDestroyMe(arg)
        #---EXIT---
        elif self.state is 'EXIT':
            pass

    def isInstalled(self, arg):
        if self.is_installed is None:
            self.is_installed = dhninit.check_install() 
        return self.is_installed
    
    def isGUIPossible(self, arg):
        if dhnio.Windows():
            return True
        if dhnio.Linux():
            return dhnio.X11_is_running()
        return False

    def doUpdate(self, arg):
        reactor.callLater(0, webcontrol.OnUpdateStartingPage)

    def doInitLocal(self, arg):
        maybeDeferred(dhninit.init_local, arg).addCallback(
            lambda x: self.automat('init-local-done'))

    def doInitContacts(self, arg):
        dhninit.init_contacts(
            lambda x: self.automat('init-contacts-done'),
            #lambda x: self.automat('init-contacts-failed'), )
            lambda x: self.automat('init-contacts-done'), )

    def doInitConnection(self, arg):
        dhninit.init_connection()

    def doInitModules(self, arg):
        maybeDeferred(dhninit.init_modules).addCallback(
            lambda x: self.automat('init-modules-done'))

    def doShowGUI(self, arg):
        d = webcontrol.init()
        if dhninit.UImode == 'show' or not self.is_installed: 
            d.addCallback(webcontrol.show)
        webcontrol.ready()

    def doDestroyMe(self, arg):
        global _Initializer
        del _Initializer
        _Initializer = None
        # automats.get_automats_by_index().pop(self.index)
        automat.objects().pop(self.index)








from socket import inet_aton

class ValidationError( Exception ) : pass

def validateIPAddress( addr ) :
    if type(addr) is not str :
        raise ValidationError, 'invalid ip address'
    try :
        inet_aton( addr )
    except :
        raise ValidationError, 'invalid ip address'

def validatePortNumber( port ) :
    if type(port) is not int :
        raise ValidationError, 'invalid port number'
    if not (0 < port < 65536) :
        raise ValidationError, 'invalid port number'

def validateInetAddress( addr ) :
    if type(addr) is not tuple :
        raise ValidationError, 'invalid inet address'
    if len(addr) != 2 :
        raise ValidationError, 'invalid inet address'
    validateIPAddress( addr[0] )
    validatePortNumber( addr[1] )



# service.py - CSpace Service
# Utilities
from bisect import bisect
import StringIO
import logging, sys
# PID, path
import os, os.path
import atexit

logger = logging.getLogger('cspace.main.service')

# Local command XMLRPC server
from SimpleXMLRPCServer import SimpleXMLRPCServer

# Select socket event dispatcher
from nitro.selectreactor import SelectReactor, SelectStoppedError
from nitro.async import AsyncOp
from nitro.http import HttpRequest

# RSA Key management
from ncrypt0.rsa import RSAKey, RSAError

# Event dispatcher
from cspace.util.eventer import Eventer
# Session management
from cspace.main.session import UserSession
# Profile management
from cspace.main.profile import listProfiles, loadProfile, saveProfileContacts, Contact
# from cspace.main.app import Reconnector, LogFile
from cspace.main.common import isValidUserName, localSettings
from cspace.main.appletserver import AppletServer, ActionConfig
from cspace.util.hexcode import hexEncode


class Reconnector( object ) :
    def __init__( self, profile, timerCallback, reactor ) :
        self.reconnecting = False
        self.profile = profile
        self.timerCallback = timerCallback 
        self.reactor = reactor
        self.timerOp = None
        self.timeLeft = 0

    def startReconnectTimer( self, errorMsg ) :
        self.errorMsg = errorMsg
        self.reconnecting = True
        self.timeLeft = 20
        self.timerOp = self.reactor.addTimer( 1, self._onTimer )

    def shutdown( self ) :
        if self.timerOp :
            self.timerOp.cancel()
            self.timerOp = None
        self.reconnecting = False

    def _onTimer( self ) :
        self.timeLeft -= 1
        if self.timeLeft == 0 :
            self.shutdown()
        self.timerCallback()


class LogFile( object ) :
    def __init__( self, settings=None, appName="CSpace" ) :
        if settings is None:
            settings = localSettings()
        configDir = settings.getConfigDir()
        logFile = os.path.join( configDir, '%s.log'%appName )
        try :
            if os.path.getsize(logFile) >= 1024*1024 :
                os.remove( logFile )
        except OSError :
            pass
        self.f = file( logFile, 'a' )

    def write( self, s ) :
        self.f.write( s )
        self.f.flush()

    def flush( self ) :
        pass


##########
## Contact action management
##
class ActionManager:
    def __init__(self):
        global logger
        self.nextActionId = 0
        self.actions = []
        self.defaultActionId = -1
        self.activeActions = ActionConfig().listActiveActions()

    def registerAction( self, action, actionCallback, actionOrder ):
        actionId = self.nextActionId
        self.nextActionId += 1
        actionDir = [adir for adir, aname, acmd, apri in self.activeActions
                          if aname == action ][0]
        info = [actionOrder,action,actionId,actionCallback,actionDir]
        pos = bisect( self.actions, info )
        self.actions.insert( pos, info )
        return actionId

    def setDefaultAction( self, actionId ) :
        if self.defaultActionId >= 0 : return False
        self.defaultActionId = actionId
        return True

    def unregisterAction( self, actionId ) :
        for pos,info in enumerate(self.actions) :
            if info[2] == actionId :
                del self.actions[pos]
                if self.defaultActionId == actionId :
                    self.defaultActionId = -1
                return True
        assert False

    def execAction( self, actionDir, contactName ) :
        global logger
        for info in self.actions :
            if info[4] == actionDir :
                # logger.info("Executing action %s on contact %s, callback id is %d" % (actionDir, contactName, id(info[3])))
                info[3]( contactName )
                return True
            else:
                logger.debug( "Not action %s" % str(info[1]) )
        return False, "Action %s not found." % actionDir

    def execDefaultAction( self, contactName ) :
        for info in self.actions :
            if info[2] is self.defaultActionId :
                info[3]( contactName )
                return


##########
## Auto-contact status prober
##
class ContactStatusChecker:
    LONGTIME = 2 * 60
    SHORTTIME = 1

    ONLINE = UserSession.ONLINE
    OFFLINE = UserSession.OFFLINE
    UNKNOWN = -1

    def __init__(self, service, reactor):
        self.service = service
        self.reactor = reactor
        self.dispatcher = self.service.dispatcher
        self.dispIdAdd = None
        self.dispIdRem = None
        self.probeOps = {}

    def startProbing(self):
        self.contacts = self.service.profile.contactNames.keys()
        self.dispIdRem = self.dispatcher.register( "contact.remove",
                                                    self.onContactRemoved )
        self.dispIdAdd = self.dispatcher.register( "contact.add",
                                                    self.onContactAdded )
        self.nextOp = self.reactor.addTimer(self.LONGTIME, self.probeContactsStatus)
        self.probeContactsStatus()

    def onContactRemoved(self, contact):
        if self.probeOps.has_key(contact.name):
            del self.probeOps[contact.name]

    def onContactAdded(self, contact):
        self.contacts.append(contact.name)
        self.probeContactStatus(contact.name)

    def stopProbing(self):
        if self.dispIdAdd:
            self.service.dispatcher.remove( self.dispIdAdd )
        if self.dispIdRem:
            self.service.dispatcher.remove( self.dispIdRem )
        self.probeOps = {}
        if self.nextOp:
            self.nextOp.cancel()

    def probeContactsStatus(self):
        if self.service.status != self.ONLINE:
            return

        self.dispatcher.trigger( "contacts.probing", self.contacts )

        for contactToProbe in self.service.profile.contactNames:
            self.probeContactStatus(contactToProbe)

    def probeContactStatus(self, cname):
        contact = self.service.profile.getContactByName(cname)
        keyID = contact.publicKey
        # logger.debug("Probing status of contact %s..." % contact.name)
        self.service.session.probeUserOnline(keyID,
                    lambda s: self.onStatusProbed(cname, s) )
        self.probeOps[cname] = True

    def onStatusProbed(self, cname, status, forceNotify=True):
        contact = self.service.profile.getContactByName(cname)
        if contact:
            oldstatus = hasattr(contact, 'status') and contact.status or self.UNKNOWN
            if status:
                contact.status = self.ONLINE
            else:
                contact.status = self.OFFLINE

            if status != oldstatus or forceNotify:
                if contact.status == self.ONLINE:
                    self.service.dispatcher.trigger("contact.online", contact)
                elif contact.status == self.OFFLINE:
                    self.service.dispatcher.trigger("contact.offline", contact)

            # logger.debug("Contact %s status: %s" % (contact.name, self.service.STATII[contact.status]))
        else:
            logger.debug("Contact removed before status returned.")

        if self.probeOps.has_key(cname):
            del self.probeOps[cname]
        if len(self.probeOps) == 0:
            self.dispatcher.trigger( "contacts.probed", self.contacts )


class CSpaceService:
    OFFLINE = UserSession.OFFLINE
    CONNECTING = UserSession.CONNECTING
    ONLINE = UserSession.ONLINE
    DISCONNECTING = UserSession.DISCONNECTING
    UNKNOWN = -1

    STATII = {
        OFFLINE: "Offline",
        CONNECTING: "Connecting",
        ONLINE: "Online",
        DISCONNECTING: "Disconnecting",
        UNKNOWN: "Unknown"
        }


    ##########
    ## Service initialization
    ##
    def __init__(self, seedNodes, settings=None, reactor=None):
        global logger
        logger.info('CSpaceService init')
        if reactor is None:
            reactor = SelectReactor()

        if settings is None:
            settings = localSettings()

        self.reactor = reactor
        self.settings = settings
        self.dispatcher = Eventer()
        self.seedNodes = seedNodes

        # Session Manager
        self.session = UserSession( self.seedNodes, reactor )

        # Our user profile
        self.profile = None

        # Session status State Machine
        self.sm = self.session.sm
        # Make sure we handle the edges/state transitions
        self.sm.insertCallback( self._onStateChange )
        # Manage actions
        self.actionManager = ActionManager()
        # Loads applets in subprocesses
        self.appletServer = AppletServer( self.session, self.actionManager,
                                          self.reactor )

        # Our contacts' statii
        self.statusProbe = ContactStatusChecker(self, reactor)

        # Our current status
        self.status = self.sm.current()
        self.statusmsg = self.STATII[self.status]
        self.exitAfterOffline = False

        self.reactor.callLater(0, self._onStarted)

    def _onStarted(self):
        self.dispatcher.trigger("service.start")

    def _error(self, msg):
        logger.error(msg)
        self.dispatcher.trigger("error", msg)
        return (False, msg)

    def run(self):
        self.reactor.run()
        try:
            self.appletServer.bridgeThread.reactor.stop()
        except AttributeError:
            pass

    def svcstatus(self):
        return self.statusmsg

    def stop(self):
        # offline first
        if self.status != self.OFFLINE:
            self.exitAfterOffline = True
            return self.offline()

        for reactor in (self.appletServer.bridgeThread.reactor,
                        self.reactor):
            reactor.stop()

        self.dispatcher.trigger("service.stop")
        return True

    ##########
    ## Go online and offline, change profiles
    ##
    def switch(self, profile, password):
        if self.status is not self.OFFLINE:
            return self._error("Not currently offline.")

        st = self.settings
        logger.debug("Changing profile...")
        profiles = [userName for userName,keyId,entry in listProfiles()]
        if not st.getInt('Settings/RememberKey',0):
            return self._error("No key found.")

        if not profile in entries:
            return self._error("No profile %s found.")

        profile = loadProfile( profile, password )
        if not profile:
            return self._error("Profile couldn't be loaded.")

        st.setString("Settings/SavedProfile", profile.name)
        st.setString("Settings/SavedPassword", password)
        saveLocalSettings()

        self.dispatcher.trigger("profile.switch", profile)
        return True


    def online(self):
        global logger
        if self.status is not self.OFFLINE:
            return self._error("Not currently offline.")

        st = self.settings
        logger.debug("Going online...")
        entries = [entry for userName,keyId,entry in listProfiles()]
        if not st.getInt('Settings/RememberKey',0):
            return self._error("No key found.")

        entry = st.getString('Settings/SavedProfile')
        password = st.getString('Settings/SavedPassword')
        if not (entry and password and (entry in entries)):
            return self._error("No profile or password.")

        profile = loadProfile( entry, password )
        if not profile:
            return self._error("Profile couldn't be loaded.")

        self.profile = profile
        logger.debug("Going online as %s keyID=%s ..." % (profile.name, profile.keyId))
        self.session.goOnline( self.profile )
        self.reconnector = Reconnector( self.profile,
                self._onReconnectTimer, self.reactor )

        return True

    def _onReconnectTimer( self ) :
        assert self.reconnector
        if self.reconnector.timeLeft > 0 :
            return
        self.profile = self.reconnector.profile
        self.reconnector.shutdown()
        if self.session.sm.current() == 0:
            self.session.goOnline( self.profile )
            self.reconnector = Reconnector( self.profile,
                self._onReconnectTimer, self.reactor )
            self.dispatcher.trigger("profile.reconnecting", self.profile)

    def offline(self):
        if self.status is not self.ONLINE:
            return self._error("Not currently online.")
        self.session.goOffline()
        self.profile = None
        self.reconnector.shutdown()
        self.reconnector = None
        return True


    def _onStateChange(self):
        """ Respond to connecting/online/disconnecting/offline notices """

        self.status = self.sm.current()
        self.statusmsg = None

        # No longer online
        if self.sm.previous() == self.ONLINE:
            # Stop checking contact statii
            self.statusProbe.stopProbing()
            # Clean up any open applets
            self.appletServer.clearConnections()

        # Now offline
        if self.sm.current() == self.OFFLINE:
            profile = self.profile
            self.profile = None
            if self.reconnector:
                if self.sm.previous() == self.CONNECTING :
                    msg = 'Connect failed.'
                else :
                    msg = 'Disconnected.'
                self.reconnector.startReconnectTimer( msg )
                self.statusmsg = msg
            self.dispatcher.trigger("profile.offline", profile)
            if self.exitAfterOffline:
                self.stop()


        # Now online
        elif self.sm.current() == self.ONLINE:
            # Restart checking contact statii
            self.dispatcher.trigger("profile.online", self.profile)
            self.statusProbe.startProbing()

        # Now connecting
        elif self.sm.current() == self.CONNECTING:
            self.dispatcher.trigger("profile.connecting", self.profile)
        elif self.sm.current() == self.DISCONNECTING:
            self.dispatcher.trigger("profile.disconnecting", self.profile)

        if not self.statusmsg:
            self.statusmsg = self.STATII[self.status]

        logger.info("STATUS: " + self.statusmsg)



    ##########
    ## List, add and remove contacts
    ##
    def list(self):
        if self.status != self.ONLINE:
            return self._error("Not connected.")

        contactsByName = self.profile.contactNames
        statii = []
        STATII = CSpaceService.STATII
        for key in contactsByName:
            contact = contactsByName[key]
            if hasattr(contact, "status"):
                buddystatus = (key, contact.status, STATII[contact.status])
            else:
                buddystatus = (key, -1, STATII[-1])
            statii.append(buddystatus)

        statii.sort()
        return statii

    def add(self, cname, keyid):
        # Lookup the contact info and call _doAddContact
        if self.status != self.ONLINE:
            return self._error("Can't lookup buddy while not online.")
        httpOp = HttpRequest( self.reactor ).get(
                # 'http://cspace.in/pubkey/%s' % keyid,
                'http://identity.datahaven.net/cspacekeys/%s' % keyid,
                self._onLookupResponse )
        self._addOp = AsyncOp( self._onAddContact, httpOp.cancel )
        self._addOp.cname = cname
        return True

    def addKey(self, cname, key):
        if cname and not isValidUserName(cname) :
            return self._error("Bad username.")

        k = RSAKey()
        try :
            k.fromPEM_PublicKey( pemPublicKey )
        except RSAError :
            return self._error("Bad PEM-encoded key.")

        contact = Contact( k, cname )
        self._onAddContact( contact )


    def _onLookupResponse(self, responseCode, data):
        if responseCode != 200 :
            self._addOp.notify( None )
            return
        inp = StringIO.StringIO( data )
        name = inp.readline().strip()
        pemPublicKey = inp.read()
        if name and not isValidUserName(name) :
            self._addOp.notify( None )
            return

        k = RSAKey()
        try :
            k.fromPEM_PublicKey( pemPublicKey )
        except RSAError :
            self._addOp.notify( None )

        contact = Contact( k, self._addOp.cname )
        self._addOp.notify( contact )

    def _onAddContact(self, contact):
        del self._addOp
        if contact == None:
            # Contact lookup failed.
            return

        self.profile.addContact( contact )
        saveProfileContacts( self.profile )
        self.dispatcher.trigger('contact.add', contact)

    def remove(self, cname=None):
        if self.status != self.ONLINE:
            return self._error("Can't modify contact list while not online.")
        # Lookup the contact in profile and remove it
        contact = self.profile.getContactByName(cname)
        if contact is None:
            return self._error("No such contact.")
        self.profile.removeContact(contact)
        saveProfileContacts( self.profile )
        self.dispatcher.trigger('contact.remove', contact)
        return True


    #########
    ## Perform contact action
    ##
    def action(self, buddy, action):
        if self.status != self.ONLINE:
            return False

        actionManager = self.actionManager
        if buddy not in self.profile.contactNames:
            return self._error("Bad buddy name: %s" % buddy)

        for actionItem in actionManager.actions:
            if actionItem[4] == action:
                logger.info("Running action %s on %s" % (action, buddy))
                retVal = actionManager.execAction(action, buddy)
                self.dispatcher.trigger("contact.action",
                                        self.profile.contactNames[buddy],
                                        action, retVal)
                return retVal
            else:
                logger.info("Not action %s" % actionItem[4])

        return self._error("Bad action name: %s" % action)

    def probe(self, buddy):
        if self.status != self.ONLINE:
            return self._error("Not online.")

        if buddy not in self.profile.contactNames:
            return self._error("Bad buddy name: %s" % buddy)

        self.statusProbe.probeContactStatus(buddy)
        return True

    def contactinfo(self, buddy):
        if self.status != self.ONLINE:
            return self._error("Not online.")

        if buddy not in self.profile.contactNames:
            return self._error("Bad buddy name: %s" % buddy)

        contact = self.profile.contactNames[buddy]
        contactKey = hexEncode(contact.publicKeyData)
        if hasattr(contact, 'status'):
            contactStatus = contact.status
        else:
            contactStatus = -1
        return [ contact.name, contactKey, self.STATII[contactStatus] ]

def _parseCommandLine():
    import optparse
    # Construct the command line argument handling
    oparser = optparse.OptionParser()

    oparser.add_option("-n", "--node", dest="dhtnode",
                        help="connect to DHT node at address", metavar="DHTIP")
    oparser.add_option("-p", "--port", dest="dhtport", type="int",
                        help="connect to DHT node on port", metavar="DHTPORT")
    oparser.add_option("-l", "--local", dest="localdht", action="store_true",
                        help="connect to DHT on localhost:10001")
    oparser.add_option("-x", "--xmlrpc", dest="rpcport", metavar="XMLPORT",
                        help="listen for local XMLRPC commands on port",
                        type="int")
    oparser.add_option("-d", "--debug", dest="debug", action="store_true",
                        help="don't redirect stderr/stdout")

	# I started a "CSpaceNetwork.py -b 208.78.96.185" command on our "datahaven.net" machine
	# So CSpace users may use it for some time.
	# Would be nice if some one else do the same to have more seed nodes
    oparser.set_default('dhtnode', '208.78.96.185') # "210.210.1.102" - that is old "dead" cspace,in server
    oparser.set_default('dhtport', 10001)
    oparser.set_default('debug', False)
    oparser.set_default('xmlport', 0)
    oparser.set_default('localdht', False)

    # Parse the command line
    (options, args) = oparser.parse_args()


    if options.localdht == True:
        from cspace.network import localip
        localip.USE_LOCALHOST = True
        options.dhtnode = "127.0.0.1"
        options.dhtport = 10001

    options.dhtport = int(options.dhtport)
    options.xmlport = int(options.xmlport)

    return options, args

def getPIDpath(appName="CSpace"):
    pidfile = "%s.run" % appName
    pidpath = os.path.join("~", ".CSpace")
    pidpath = os.path.expanduser(pidpath)
    if not os.path.exists(pidpath):
        os.makedirs(pidpath)
    pidpath = os.path.join(pidpath, pidfile)
    return pidpath

def _writePID(xmlrpcport, appletport):
    pidpath = getPIDpath()
    pidfile = file(pidpath, "wb")
    pidfile.write("%i\n%i\n%i\n" % (os.getpid(), xmlrpcport, appletport))
    pidfile.close()

def _deletePID():
    pidpath = getPIDpath()
    os.unlink(pidpath)

def main():
    global logger, reactor
    options, args = _parseCommandLine()
    settings = localSettings()

    if not options.debug:
        sys.stdout = sys.stderr = LogFile( settings )
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.DEBUG)
    logger.addHandler( logging.StreamHandler() )
    logging.getLogger('nitro.selectreactor').addHandler( logging.StreamHandler() )

    reactor = SelectReactor()
    seedNodes = [(options.dhtnode, options.dhtport)]
    service = CSpaceService(seedNodes, settings, reactor)
    server = SimpleXMLRPCServer(('localhost', options.xmlport))
    reactor.addReadCallback(server, server.handle_request)
    server.register_instance(service, allow_dotted_names = True)

    xmlrpcport = server.socket.getsockname()[1]
    logger.info("rpcserver listenport = %i" % xmlrpcport)
    appletport = service.appletServer.listenPort
    logger.info("seed nodes = (%s:%i)" % (options.dhtnode, options.dhtport))
    _writePID(xmlrpcport, appletport)
    service.run()
    _deletePID()
    logger.info("server stopped.")
